"""
The wntr.graphics.network module includes methods plot the
water network model.
"""
import logging
import math
import networkx as nx
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.path as mpath
from matplotlib import animation
import matplotlib as mpl
import numpy as np

try:
    import plotly
except:
    plotly = None
try:
    import folium
except:
    folium = None
    
from wntr.graphics.color import custom_colormap

logger = logging.getLogger(__name__)


arrow_verts = [
    (0.0, 0.0),
    (0.5, 0.5),
    (0.5, -0.5),
    (0.0, 0.0),
]

arrow_marker = mpath.Path(arrow_verts)

def _get_angle(line, loc=0.5):
    # calculate orientation angle
    p1 = line.interpolate(loc-0.01, normalized=True)
    p2 = line.interpolate(loc+0.01, normalized=True)
    angle = math.atan2(p2.y-p1.y, p2.x - p1.x) # radians
    angle = math.degrees(angle)
    return angle

def _format_node_attribute(node_attribute, wn):
    
    if isinstance(node_attribute, str):
        node_attribute = wn.query_node_attribute(node_attribute)
    if isinstance(node_attribute, list):
        node_attribute = dict(zip(node_attribute,[1]*len(node_attribute)))
    if isinstance(node_attribute, pd.Series):
        node_attribute = dict(node_attribute)
    
    return node_attribute

def _format_link_attribute(link_attribute, wn):
    
    if isinstance(link_attribute, str):
        link_attribute = wn.query_link_attribute(link_attribute)
    if isinstance(link_attribute, list):
        link_attribute = dict(zip(link_attribute,[1]*len(link_attribute)))
    if isinstance(link_attribute, pd.Series):
        link_attribute = dict(link_attribute)
            
    return link_attribute


def plot_network(wn, node_attribute=None, link_attribute=None, title=None,
               node_size=20, node_range=[None,None], node_alpha=1, node_cmap=None, node_labels=False,
               link_width=1, link_range=[None,None], link_alpha=1, link_cmap=None, link_labels=False,
               add_colorbar=True, node_colorbar_label='Node', link_colorbar_label='Link', 
               directed=False, legend=False, ax=None, show_plot=True, filename=None, backend='nx'):
    """
    Plot network graphic
	
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
		
    node_attribute : None, str, list, pd.Series, or dict, optional
	
        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float
    
	link_attribute : None, str, list, pd.Series, or dict, optional
	
        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.
		  
    title: str, optional
        Plot title 

    node_size: int, optional
        Node size 

    node_range: list, optional
        Node color range ([None,None] indicates autoscale)
        
    node_alpha: int, optional
        Node transparency
        
    node_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width: int, optional
        Link width
		
    link_range : list, optional
        Link color range ([None,None] indicates autoscale)
		
    link_alpha : int, optional
        Link transparency
    
    link_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name.
        
    add_colorbar: bool, optional
        Add colorbar

    node_colorbar_label: str, optional
        Node colorbar label
        
    link_colorbar_label: str, optional
        Link colorbar label
        
    directed: bool, optional
        If True, plot the directed graph
        
    legend: bool, optional
        If True, show legend for component types (Tanks, Reservoirs, etc)
        Only available for backend='gpd'
    
    ax: matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
    
    show_plot: bool, optional
        If True, show plot with plt.show()
    
    filename : str, optional
        Filename used to save the figure
        
    backend: str, optional
        Backend used for plotting: 
        'nx' for networkx (default) 
        'gpd' for geopandas. Adds icons for components and vertices on links
        
    Returns
    -------
    ax : matplotlib axes object  
    """
    if backend == 'nx':
        ax = _plot_network_nx(wn=wn, node_attribute=node_attribute, link_attribute=link_attribute, title=title,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, node_colorbar_label=node_colorbar_label, link_colorbar_label=link_colorbar_label, 
               directed=directed, ax=ax, show_plot=show_plot, filename=filename)
    elif backend == 'gpd':
        ax = _plot_network_gpd(wn=wn, node_attribute=node_attribute, link_attribute=link_attribute, title=title,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, node_colorbar_label=node_colorbar_label, link_colorbar_label=link_colorbar_label, 
               directed=directed, legend=legend, ax=ax, show_plot=show_plot, filename=filename)
    else:
        raise Exception("Backend choice {backend} unrecognized. Please use 'nx' for networkx or 'gpd' for geopandas.")

    return ax

def _plot_network_nx(wn, node_attribute=None, link_attribute=None, title=None,
               node_size=20, node_range=[None,None], node_alpha=1, node_cmap=None, node_labels=False,
               link_width=1, link_range=[None,None], link_alpha=1, link_cmap=None, link_labels=False,
               add_colorbar=True, node_colorbar_label='Node', link_colorbar_label='Link', 
               directed=False, ax=None, show_plot=True, filename=None):
    """
    Plot network graphic
	
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
		
    node_attribute : None, str, list, pd.Series, or dict, optional
	
        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float
    
	link_attribute : None, str, list, pd.Series, or dict, optional
	
        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.
		  
    title: str, optional
        Plot title 

    node_size: int, optional
        Node size 

    node_range: list, optional
        Node color range ([None,None] indicates autoscale)
        
    node_alpha: int, optional
        Node transparency
        
    node_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width: int, optional
        Link width
		
    link_range : list, optional
        Link color range ([None,None] indicates autoscale)
		
    link_alpha : int, optional
        Link transparency
    
    link_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name.
        
    add_colorbar: bool, optional
        Add colorbar

    node_colorbar_label: str, optional
        Node colorbar label
        
    link_colorbar_label: str, optional
        Link colorbar label
        
    directed: bool, optional
        If True, plot the directed graph
    
    ax: matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)
    
    show_plot: bool, optional
        If True, show plot with plt.show()
    
    filename : str, optional
        Filename used to save the figure
        
    Returns
    -------
    ax : matplotlib axes object  
    """
    
    def _format_node_attribute(node_attribute, wn):
    
        if isinstance(node_attribute, str):
            node_attribute = wn.query_node_attribute(node_attribute)
        if isinstance(node_attribute, list):
            node_attribute = dict(zip(node_attribute,[1]*len(node_attribute)))
        if isinstance(node_attribute, pd.Series):
            node_attribute = dict(node_attribute)
        
        return node_attribute

    def _format_link_attribute(link_attribute, wn):
        
        if isinstance(link_attribute, str):
            link_attribute = wn.query_link_attribute(link_attribute)
        if isinstance(link_attribute, list):
            link_attribute = dict(zip(link_attribute,[1]*len(link_attribute)))
        if isinstance(link_attribute, pd.Series):
            link_attribute = dict(link_attribute)
                
        return link_attribute

    if ax is None: # create a new figure
        plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
        
    # Graph
    G = wn.to_graph()
    if not directed:
        G = G.to_undirected()

    # Position
    pos = nx.get_node_attributes(G,'pos')
    if len(pos) == 0:
        pos = None

    # Define node properties
    add_node_colorbar = add_colorbar
    if node_attribute is not None:
        
        if isinstance(node_attribute, list):
            if node_cmap is None:
                node_cmap = ['red', 'red']
            add_node_colorbar = False
        
        if node_cmap is None:
            node_cmap = plt.get_cmap('Spectral_r')
        elif isinstance(node_cmap, list):
            if len(node_cmap) == 1:
                node_cmap = node_cmap*2
            node_cmap = custom_colormap(len(node_cmap), node_cmap)  
         
        node_attribute = _format_node_attribute(node_attribute, wn)
        nodelist,nodecolor = zip(*node_attribute.items())

    else:
        nodelist = None
        nodecolor = 'k'
    
    add_link_colorbar = add_colorbar
    if link_attribute is not None:
        
        if isinstance(link_attribute, list):
            if link_cmap is None:
                link_cmap = ['red', 'red']
            add_link_colorbar = False

        if link_cmap is None:
            link_cmap = plt.get_cmap('Spectral_r')
        elif isinstance(link_cmap, list):
            if len(link_cmap) == 1:
                link_cmap = link_cmap*2
            link_cmap = custom_colormap(len(link_cmap), link_cmap)  
            
        link_attribute = _format_link_attribute(link_attribute, wn)
        
        # Replace link_attribute dictionary defined as
        # {link_name: attr} with {(start_node, end_node, link_name): attr}
        attr = {}
        for link_name, value in link_attribute.items():
            link = wn.get_link(link_name)
            attr[(link.start_node_name, link.end_node_name, link_name)] = value
        link_attribute = attr
        
        linklist,linkcolor = zip(*link_attribute.items())
    else:
        linklist = None
        linkcolor = 'k'
    
    if title is not None:
        ax.set_title(title)
        
    edge_background = nx.draw_networkx_edges(G, pos, edge_color='grey', 
                                             width=0.5, ax=ax)
    
    nodes = nx.draw_networkx_nodes(G, pos, 
            nodelist=nodelist, node_color=nodecolor, node_size=node_size, 
            alpha=node_alpha, cmap=node_cmap, vmin=node_range[0], vmax = node_range[1], 
            linewidths=0, ax=ax)
    edges = nx.draw_networkx_edges(G, pos, edgelist=linklist, arrows=directed,
            edge_color=linkcolor, width=link_width, alpha=link_alpha, edge_cmap=link_cmap, 
            edge_vmin=link_range[0], edge_vmax=link_range[1], ax=ax)
    if node_labels:
        labels = dict(zip(wn.node_name_list, wn.node_name_list))
        nx.draw_networkx_labels(G, pos, labels, font_size=7, ax=ax)
    if link_labels:
        labels = {}
        for link_name in wn.link_name_list:
            link = wn.get_link(link_name)
            labels[(link.start_node_name, link.end_node_name)] = link_name
        nx.draw_networkx_edge_labels(G, pos, labels, font_size=7, ax=ax)
    if add_node_colorbar and node_attribute:
        clb = plt.colorbar(nodes, shrink=0.5, pad=0, ax=ax)
        clb.ax.set_title(node_colorbar_label, fontsize=10)
    if add_link_colorbar and link_attribute:
        if link_range[0] is None:
            vmin = min(link_attribute.values())
        else:
            vmin = link_range[0]
        if link_range[1] is None:
            vmax = max(link_attribute.values())
        else:
            vmax = link_range[1]
        sm = plt.cm.ScalarMappable(cmap=link_cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))
        sm.set_array([])
        clb = plt.colorbar(sm, shrink=0.5, pad=0.05, ax=ax)
        clb.ax.set_title(link_colorbar_label, fontsize=10)
        
    ax.axis('off')
    
    if filename:
        plt.savefig(filename)
    
    if show_plot is True:
        plt.show(block=False)
    
    return ax

def _plot_network_gpd(
    wn, node_attribute=None, link_attribute=None, title=None,
    node_size=20, node_range=None, node_alpha=1, node_cmap=None, node_labels=False,
    link_width=1, link_range=None, link_alpha=1, link_cmap=None, link_labels=False,
    add_colorbar=True, node_colorbar_label=None, link_colorbar_label=None, 
    directed=False, legend=False, ax=None, show_plot=True, filename=None):
    """
    Plot network graphic
	
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
		
    node_attribute : None, str, list, pd.Series, or dict, optional
	
        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float
    
	link_attribute : None, str, list, pd.Series, or dict, optional
	
        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.
		  
    title: str, optional
        Plot title 

    node_size: int, optional
        Node size 

    node_range: list, optional
        Node color range ([None,None] indicates autoscale)
        
    node_alpha: int, optional
        Node transparency
        
    node_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width: int, optional
        Link width
		
    link_range : list, optional
        Link color range ([None,None] indicates autoscale)
		
    link_alpha : int, optional
        Link transparency
    
    link_cmap: matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name.
        
    add_colorbar: bool, optional
        Add colorbar

    node_colorbar_label: str, optional
        Node colorbar label
        
    link_colorbar_label: str, optional
        Link colorbar label
        
    directed: bool, optional
        If True, plot the directed graph
    
    ax: matplotlib axes object, optional
        Axes for plotting (None indicates that a new figure with a single 
        axes will be used)

    show_plot: bool, optional
        If True, show plot with plt.show()
    
    filename : str, optional
        Filename used to save the figure
        
    Returns
    -------
    ax : matplotlib axes object  
    """
    if ax is None: # create a new figure
        plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
        
    if title is not None:
        ax.set_title(title)
    
    aspect = "equal"
    
    tank_marker = "D"
    reservoir_marker = "s"
    
    if link_cmap is None:
        link_cmap = plt.get_cmap('Spectral_r')
    if node_cmap is None:
        node_cmap = plt.get_cmap('Spectral_r')
        
    if link_range is None:
        link_range = (None, None)
    if node_range is None:
        node_range = (None, None)
    
    # use attribute name if no other label is provided
    if node_colorbar_label is None and isinstance(node_attribute, str):
        node_colorbar_label = node_attribute
    if link_colorbar_label is None and isinstance(link_attribute, str):
        link_colorbar_label = link_attribute 
        
    wn_gis = wn.to_gis()
    # add node_type so that node assets can be plotted separately
    wn_gis.junctions["node_type"] = "Junction"
    wn_gis.tanks["node_type"] = "Tank"
    wn_gis.reservoirs["node_type"] = "Reservoir"
    link_gdf = pd.concat((wn_gis.pipes, wn_gis.pumps, wn_gis.valves))
    node_gdf = pd.concat((wn_gis.junctions, wn_gis.tanks, wn_gis.reservoirs))
    
    # Node attribute
    node_kwds = {}
    node_cbar = add_colorbar
    if node_attribute is not None:
        node_gdf["_attribute"] = _format_node_attribute(node_attribute, wn)
        node_kwds["column"] = "_attribute"
        
        # handle cbar/cmap
        if isinstance(node_attribute, list):
            node_kwds["cmap"] = custom_colormap(2,["red", "red"])
            node_cbar = False
        elif isinstance(node_attribute, (dict, pd.Series, str)):
            node_kwds["cmap"] = node_cmap
            
            # manually extract min/max if no range is given
            node_attribute_values = node_gdf[node_kwds["column"]]
            if node_range[0] is None:
                node_kwds["vmin"] = np.nanmin(node_attribute_values)
            else:
                node_kwds["vmin"] = node_range[0]
            if node_range[1] is None:
                node_kwds["vmax"] = np.nanmax(node_attribute_values)
            else:
                node_kwds["vmax"] = node_range[1]
        else:
            raise TypeError("attribute must be dict, Series, list, or str")
    else:
        node_kwds["color"] = "black"
        node_cbar = False
        
    node_kwds["alpha"] = node_alpha
    node_kwds["markersize"] = node_size
    
    node_cbar_kwds = {}
    node_cbar_kwds["shrink"] = 0.5
    node_cbar_kwds["pad"] = 0.0
    node_cbar_kwds["alpha"] = node_alpha
    node_cbar_kwds["label"] = node_colorbar_label
    
    # Link attribute
    link_kwds = {}
    link_cbar = add_colorbar
    if link_attribute is not None:
        link_gdf["_attribute"] = pd.Series(_format_link_attribute(link_attribute, wn))
        link_kwds["column"] = "_attribute"
        
        # handle cbar/cmap
        if isinstance(link_attribute, list):
            link_kwds["cmap"] = custom_colormap(2,["red", "red"])
            link_cbar = False
        elif isinstance(link_attribute, (dict, pd.Series, str)):
            link_kwds["cmap"] = link_cmap
            
            # manually extract min/max if no range is given
            link_attribute_values = link_gdf[link_kwds["column"]]
            if link_range[0] is None:
                link_kwds["vmin"] = np.nanmin(link_attribute_values)
            else:
                link_kwds["vmin"] = link_range[0]
            if link_range[1] is None:
                link_kwds["vmax"] = np.nanmax(link_attribute_values)
            else:
                link_kwds["vmax"] = link_range[1]
        else:
            raise TypeError("attribute must be dict, Series, list, or str")
    else:
        link_kwds["color"] = "black"
        link_cbar = False
    
    link_kwds["linewidth"] = link_width
    link_kwds["alpha"] = link_alpha
    
    background_link_kwds = {}
    background_link_kwds["color"] = "grey"
    background_link_kwds["linewidth"] = link_width / 2
    background_link_kwds["alpha"] = link_alpha
    
    link_cbar_kwds = {}
    link_cbar_kwds["shrink"] = 0.5
    link_cbar_kwds["pad"] = 0.05
    link_cbar_kwds["label"] = link_colorbar_label
    link_cbar_kwds["alpha"] = link_alpha
    
    missing_node_kwds={"color": "black"}
    missing_link_kwds={"color": "black"}

    # plot nodes - each type is plotted separately to allow for different marker types
    node_gdf[node_gdf.node_type == "Junction"].plot(
        ax=ax, aspect=aspect, zorder=3, legend=False, label="Junction", missing_kwds=missing_node_kwds, **node_kwds)
    
    node_kwds["markersize"] = node_size * 2.0
    node_gdf[node_gdf.node_type == "Tank"].plot(
        ax=ax, aspect=aspect, zorder=4, marker=tank_marker, legend=False, label="Tank", missing_kwds=missing_node_kwds, **node_kwds)
    
    node_kwds["markersize"] = node_size * 3.0
    node_gdf[node_gdf.node_type == "Reservoir"].plot(
        ax=ax, aspect=aspect, zorder=5, marker=reservoir_marker, legend=False, label="Reservoir", missing_kwds=missing_node_kwds,**node_kwds)
    
    if node_cbar:
        sm = mpl.cm.ScalarMappable(cmap=node_kwds["cmap"])
        sm.set_clim(node_kwds["vmin"], node_kwds["vmax"])

        node_cbar = ax.figure.colorbar(sm, ax=ax, **node_cbar_kwds)
    
    # plot links
    # background
    link_gdf.plot(
        ax=ax, aspect=aspect, zorder=1, legend=False, **background_link_kwds)
    
    # main plot
    link_gdf.plot(
        ax=ax, aspect=aspect, zorder=2, legend=False, missing_kwds=missing_link_kwds, **link_kwds)
    
    if link_cbar:
        sm = mpl.cm.ScalarMappable(cmap=link_kwds["cmap"])
        sm.set_clim(link_kwds["vmin"], link_kwds["vmax"])

        link_cbar = ax.figure.colorbar(sm, ax=ax, **link_cbar_kwds)

    if node_labels:
        for x, y, label in zip(node_gdf.geometry.x, node_gdf.geometry.y, node_gdf.index):
            ax.annotate(label, xy=(x, y))#, xytext=(3, 3),)# textcoords="offset points")
            
    if link_labels:
        midpoints = link_gdf.geometry.apply(lambda x: x.interpolate(0.5, normalized=True))
        for x, y, label in zip(midpoints.geometry.x, midpoints.geometry.y, link_gdf.index):
            ax.annotate(label, xy=(x, y))#, xytext=(3, 3),)# textcoords="offset points") 
            
    if directed:
        link_gdf["_midpoint"] = link_gdf.geometry.interpolate(0.5, normalized=True)
        link_gdf["_angle"] = link_gdf.apply(lambda row: _get_angle(row.geometry), axis=1)
        for idx , row in link_gdf.iterrows():
            x,y = row["_midpoint"].x, row["_midpoint"].y
            angle = row["_angle"]
            ax.scatter(x,y, color="black", s=50, marker=(3,0, angle-90))

    if legend:
        handles, labels = ax.get_legend_handles_labels()
        leg = ax.legend(handles, labels, loc='upper right', title="Legend")
    
    ax.axis('off')
    
    if filename:
        plt.savefig(filename)
    
    if show_plot is True:
        plt.show(block=False)
    
    return ax

def plot_interactive_network(wn, node_attribute=None, node_attribute_name = 'Value', title=None,
               node_size=8, node_range=[None,None], node_cmap='Jet', node_labels=True,
               link_width=1, add_colorbar=True, reverse_colormap=False,
               figsize=[700, 450], round_ndigits=2, add_to_node_popup=None, 
               filename='plotly_network.html', auto_open=True):
    """
    Create an interactive scalable network graphic using plotly. 

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
    node_attribute : None, str, list, pd.Series, or dict, optional
        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float.
          The time index is not used in the plot.
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float
    node_attribute_name : str, optional 
        The node attribute name, which is used in the node popup and node legend
    title : str, optional
        Plot title
    node_size : int, optional
        Node size
    node_range : list, optional
        Node range ([None,None] indicates autoscale)
    node_cmap : palette name string, optional
        Node colormap, options include Greys, YlGnBu, Greens, YlOrRd, Bluered, 
        RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, Blackbody, 
        Earth, Electric, Viridis
    
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name and 
        attribute value.
        
    link_width : int, optional
        Link width
    
    add_colorbar : bool, optional
        Add colorbar
        
    reverse_colormap : bool, optional
        Reverse colormap
        
    figsize: list, optional
        Figure size in pixels
    round_ndigits : int, optional
        Number of digits to round node values used in the label
    
    add_to_node_popup : None or pd.DataFrame, optional
        To add additional information to the node popup, use a DataFrame with 
        node name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given node.
        
    filename : string, optional
        HTML file name
    
    auto_open : bool, optional
        Open the HTML file after creation
    """
    if plotly is None:
        raise ImportError('plotly is required')
        
    # Graph
    G = wn.to_graph()
    
    # Node attribute
    if node_attribute is not None:
        if isinstance(node_attribute, list):
            node_cmap = 'Reds'
            add_colorbar = False
        node_attribute = _format_node_attribute(node_attribute, wn)
    else:
        add_colorbar = False
        
    # Create edge trace
    edge_trace = plotly.graph_objs.Scatter(
        x=[], 
        y=[], 
        text=[],
        hoverinfo='text',
        mode='lines',
        line=dict(
            #colorscale=link_cmap,
            #reversescale=reverse_colormap,
            color='#888', #[], 
            width=link_width))
    for edge in G.edges():
        x0, y0 = G.nodes[edge[0]]['pos']
        x1, y1 = G.nodes[edge[1]]['pos']
        edge_trace['x'] += tuple([x0, x1, None])
        edge_trace['y'] += tuple([y0, y1, None])
#        try:
#            # Add link attributes
#            link_name = G[edge[0]][edge[1]].keys()[0]
#            edge_trace['line']['color'] += tuple([pipe_attr[link_name]])
#            edge_info = 'Edge ' + str(link_name)
#            edge_trace['text'] += tuple([edge_info])
#        except:
#            pass
#    edge_trace['colorbar']['title'] = 'Link colorbar title'
    
    # Create node trace      
    node_trace = plotly.graph_objs.Scatter(
        x=[], 
        y=[], 
        text=[],
        hoverinfo='text',
        mode='markers', 
        marker=dict(
            showscale=add_colorbar,
            colorscale=node_cmap, 
            cmin=node_range[0], # TODO: Not sure this works
            cmax=node_range[1], # TODO: Not sure this works
            reversescale=reverse_colormap,
            color=[], 
            size=node_size,         
            #opacity=0.75,
            colorbar=dict(
                thickness=15,
                xanchor='left',
                titleside='right'),
            line=dict(width=1)))
    for node in G.nodes():
        x, y = G.nodes[node]['pos']
        node_trace['x'] += tuple([x])
        node_trace['y'] += tuple([y])
        try:
            # Add node attributes
            node_trace['marker']['color'] += tuple([node_attribute[node]])
            #node_trace['marker']['size'].append(node_size)

            # Add node labels
            if node_labels:
                node_info = wn.get_node(node).node_type + ': ' + str(node) + '<br>'+ \
                            node_attribute_name + ': ' + str(round(node_attribute[node],round_ndigits))
                if add_to_node_popup is not None:
                    if node in add_to_node_popup.index:
                        for key, val in add_to_node_popup.loc[node].iteritems():
                            node_info = node_info + '<br>' + \
                                key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                            
                node_trace['text'] += tuple([node_info])
        except:
            node_trace['marker']['color'] += tuple(['#888'])
            if node_labels:
                node_info = wn.get_node(node).node_type + ': ' + str(node)
                
                node_trace['text'] += tuple([node_info])
            #node_trace['marker']['size'] += tuple([5])
    #node_trace['marker']['colorbar']['title'] = 'Node colorbar title'
    
    # Create figure
    data = [edge_trace, node_trace]
    layout = plotly.graph_objs.Layout(
                    title=title,
                    titlefont=dict(size=16),
                    showlegend=False, 
                    width=figsize[0],
                    height=figsize[1],
                    hovermode='closest',
                    margin=dict(b=20,l=5,r=5,t=40),
                    xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                    yaxis=dict(showgrid=False, zeroline=False, showticklabels=False))
    
    fig = plotly.graph_objs.Figure(data=data,layout=layout)
    if filename:
        plotly.offline.plot(fig, filename=filename, auto_open=auto_open)  
    else:
        plotly.offline.plot(fig, auto_open=auto_open)  

def plot_leaflet_network(wn, node_attribute=None, link_attribute=None, 
               node_attribute_name = 'Value', 
               link_attribute_name = 'Value',
               node_size=2, node_range=[None,None], 
               node_cmap=['cornflowerblue', 'forestgreen', 'gold', 'firebrick'], 
               node_cmap_bins = 'cut', node_labels=True,
               link_width=2, link_range=[None,None], 
               link_cmap=['cornflowerblue', 'forestgreen', 'gold', 'firebrick'], 
               link_cmap_bins='cut', link_labels=True,
               add_legend=False, round_ndigits=2, zoom_start=13, 
               add_to_node_popup=None, add_to_link_popup=None,
               filename='leaflet_network.html'):
    """
    Create an interactive scalable network graphic on a Leaflet map using folium.

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
    node_attribute : None, str, list, pd.Series, or dict, optional
        - If node_attribute is a string, then a node attribute dictionary is
          created using node_attribute = wn.query_node_attribute(str)
        - If node_attribute is a list, then each node in the list is given a 
          value of 1.
        - If node_attribute is a pd.Series, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float. 
        - If node_attribute is a dict, then it should be in the format
          {nodeid: x} where nodeid is a string and x is a float
    link_attribute : None, str, list, pd.Series, or dict, optional
        - If link_attribute is a string, then a link attribute dictionary is
          created using edge_attribute = wn.query_link_attribute(str)
        - If link_attribute is a list, then each link in the list is given a 
          value of 1.
        - If link_attribute is a pd.Series, then it should be in the format
          {linkid: x} where linkid is a string and x is a float. 
        - If link_attribute is a dict, then it should be in the format
          {linkid: x} where linkid is a string and x is a float.
    node_attribute_name : str, optional 
        The node attribute name, which is used in the node popup and node legend
        
    link_attribute_name : str, optional 
        The link attribute name, which is used in the link popup and link legend
        
    node_size : int, optional
        Node size 
    node_range : list, optional
        Node range ([None,None] indicates autoscale)
    node_cmap : list of color names, optional
        Node colors 
    
    node_cmap_bins: string, optional
        Node color bins, 'cut' or 'qcut'
    
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width : int, optional
        Link width
    link_range : list, optional
        Link range ([None,None] indicates autoscale)
    link_cmap : list of color names, optional
        Link colors
    
    link_cmap_bins: string, optional
        Link color bins, 'cut' or 'qcut'
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name. 
    
    add_legend: bool, optional
         Add a legend to the map
    
    round_ndigits : int, optional
        Rounds digits in the popup
        
    zoom_start : int, optional
        Zoom start used to set initial scale of the map
    
    add_to_node_popup : None or pd.DataFrame, optional
        To add additional information to the node popup, use a DataFrame with 
        node name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given node.
        
    add_to_link_popup : None or pd.DataFrame, optional
        To add additional information to the link popup, use a DataFrame with 
        link name as index and attributes as values.  Column names will be added
        to the popup along with each value for a given link.
        
    filename : str, optional
        Filename used to save the map
    """
    
    if folium is None:
        raise ImportError('folium is required')
    
    if node_attribute is not None:
        if isinstance(node_attribute, list):
            node_cmap=['red']
        node_attribute = _format_node_attribute(node_attribute, wn)
        node_attribute = pd.Series(node_attribute)
        if node_range[0] is not None:
            node_attribute[node_attribute < node_range[0]] = node_range[0]
        if node_range[1] is not None:
            node_attribute[node_attribute > node_range[1]] = node_range[1]
        if node_cmap_bins == 'cut':
            node_colors, node_bins = pd.cut(node_attribute, len(node_cmap), 
                                                labels=node_cmap, retbins =True)
        elif node_cmap_bins == 'qcut':
            node_colors, node_bins = pd.qcut(node_attribute, len(node_cmap), 
                                                 labels=node_cmap, retbins =True)
        
    if link_attribute is not None:
        if isinstance(link_attribute, list):
            link_cmap=['red']
        link_attribute = _format_link_attribute(link_attribute, wn)
        link_attribute = pd.Series(link_attribute)
        if link_range[0] is not None:
            link_attribute[link_attribute < link_range[0]] = link_range[0]
        if link_range[1] is not None:
            link_attribute[link_attribute > link_range[1]] = link_range[1]
        if link_cmap_bins == 'cut':
            link_colors, link_bins  = pd.cut(link_attribute, len(link_cmap), 
                                             labels=link_cmap, retbins =True)
        elif link_cmap_bins == 'qcut':
            link_colors, link_bins  = pd.qcut(link_attribute, len(link_cmap), 
                                              labels=link_cmap, retbins =True)
        
    G = wn.to_graph()
    pos = nx.get_node_attributes(G,'pos')
    center = pd.DataFrame(pos).mean(axis=1)
    
    m = folium.Map(location=[center.iloc[1], center.iloc[0]], zoom_start=zoom_start, 
                   tiles='cartodbpositron')
    #folium.TileLayer('cartodbpositron').add_to(m)
    
    # Node popup
    node_popup = {k: '' for k in wn.node_name_list}
    if node_labels:
        for name, node in wn.nodes():
            node_popup[name] = node.node_type + ': ' + name
            if node_attribute is not None:
                if name in node_attribute.index:
                    node_popup[name] = node_popup[name] + '<br>' + \
                        node_attribute_name + ': ' + '{:.{prec}f}'.format(node_attribute[name], prec=round_ndigits)
            if add_to_node_popup is not None:
                if name in add_to_node_popup.index:
                    for key, val in add_to_node_popup.loc[name].iteritems():
                        node_popup[name] = node_popup[name] + '<br>' + \
                            key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                 
    # Link popup
    link_popup = {k: '' for k in wn.link_name_list}
    if link_labels:
        for name, link in wn.links():
            link_popup[name] = link.link_type + ': ' + name
            if link_attribute is not None:
                if name in link_attribute.index:
                    link_popup[name] = link_popup[name] + '<br>' + \
                        link_attribute_name + ': ' + '{:.{prec}f}'.format(link_attribute[name], prec=round_ndigits)
            if add_to_link_popup is not None:
                if name in add_to_link_popup.index:
                    for key, val in add_to_link_popup.loc[name].iteritems():
                        link_popup[name] = link_popup[name] + '<br>' + \
                            key + ': ' + '{:.{prec}f}'.format(val, prec=round_ndigits)
                            
    if node_size > 0:
        for name, node in wn.nodes():
            loc = (node.coordinates[1], node.coordinates[0])
            radius = node_size
            color = 'black'
            if node_labels:
                popup = node_popup[name]
            else:
                popup = None
                    
            if node_attribute is not None:
                if name in node_attribute.index:
                    color = node_colors[name]
                else:
                    radius = 0.1
            
            folium.CircleMarker(loc, popup=popup, color=color, fill=True, 
                                fill_color=color, radius=radius, fill_opacity=0.7, opacity=0.7).add_to(m)
            
    if link_width > 0:
        for name, link in wn.links():            
            start_loc = (link.start_node.coordinates[1], link.start_node.coordinates[0])
            end_loc = (link.end_node.coordinates[1], link.end_node.coordinates[0])
            weight = link_width
            color='black'
            if link_labels:
                popup = link_popup[name]
            else:
                popup = None
            
            if link_attribute is not None:
                if name in link_attribute.index:
                    color = link_colors[name]
                else:
                    weight = 1.5
            
            folium.PolyLine([start_loc, end_loc], popup=popup, color=color, 
                            weight=weight, opacity=0.7).add_to(m)
    
    if (add_legend) & ((len(node_cmap) >= 1) or (len(link_cmap) >= 1)):
        if node_attribute is not None:  #Produce node legend
            height = 50+len(node_cmap)*20 + (int(len(node_attribute_name)/20) + 1)*20
            node_legend_html = """<div style="position: fixed; 
        bottom: 50px; left: 50px; width: 150px; height: """+str(height)+"""px; 
        background-color:white;z-index:9999; font-size:14px; "><br>
            <b><P ALIGN=CENTER>""" + "Node Legend: " + node_attribute_name + """</b> </P>"""
            for color, val in zip(node_cmap, node_bins[0:-1]):
                val = '{:.{prec}f}'.format(val, prec=round_ndigits)
                node_legend_html += """
                &emsp;<i class="fa fa-circle fa-1x" 
                style="color:"""+ color +""" "></i> >= """+ val +""" <br>"""
            node_legend_html += """</div>"""
            m.get_root().html.add_child(folium.Element(node_legend_html))
			
        if link_attribute is not None:   #Produce link legend
            height = 50+len(link_cmap)*20 + (int(len(link_attribute_name)/20) + 1)*20
            link_legend_html = """<div style="position: fixed; 
			bottom: 50px; left: 250px; width: 150px; height: """+str(height)+"""px; 
			background-color:white;z-index:9999; font-size:14px; "><br>
            <b><P ALIGN=CENTER>""" + "Link Legend: " + link_attribute_name + """</b> </P>"""
            for color, val in zip(link_cmap, link_bins[0:-1]):
                val = '{:.{prec}f}'.format(val, prec=round_ndigits)
                link_legend_html += """
               &emsp;<i class="fa fa-minus fa-1x" 
                style="color:"""+ color +""" "></i> >= """+ val +""" <br>"""
            link_legend_html += """</div>"""
            m.get_root().html.add_child(folium.Element(link_legend_html))
    
    #plugins.Search(points, search_zoom=20, ).add_to(m)
    #if add_longlat_popup:
    #    m.add_child(folium.LatLngPopup())
    
    folium.LayerControl().add_to(m)
    
    m.save(filename)
 
def network_animation(wn, node_attribute=None, link_attribute=None, title=None,
               node_size=20, node_range=[None,None], node_alpha=1, node_cmap=None, node_labels=False,
               link_width=1, link_range=[None,None], link_alpha=1, link_cmap=None, link_labels=False,
               add_colorbar=True, directed=False, ax=None, repeat=True):
    """
    Create a network animation
    
    Parameters
    ----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
    node_attribute : pd.DataFrame, optional
        Node attributes stored in a pandas DataFrames, where the index is 
        time and columns are the node name 
    link_attribute : pd.DataFrame, optional
        Link attributes stored in a pandas DataFrames, where the index is 
        time and columns are the link name 
    title : str, optional
        Plot title 
    node_size : int, optional
        Node size 
    node_range : list, optional
        Node range ([None,None] indicates autoscale)
        
    node_alpha : int, optional
        Node transparency
        
    node_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Node colormap 
        
    node_labels: bool, optional
        If True, the graph will include each node labelled with its name. 
        
    link_width : int, optional
        Link width
    link_range : list, optional
        Link range ([None,None] indicates autoscale)
    link_alpha : int, optional
        Link transparency
    
    link_cmap : matplotlib.pyplot.cm colormap or list of named colors, optional
        Link colormap
        
    link_labels: bool, optional
        If True, the graph will include each link labelled with its name. 
        
    add_colorbar : bool, optional
        Add colorbar
    directed : bool, optional
        If True, plot the directed graph
    
    repeat : bool, optional
        If True, the animation will repeat
        
    Returns
    -------
    matplotlib animation
    """
    
    if node_attribute is not None:
        node_index = node_attribute.index
        initial_node_values = node_attribute.iloc[0, :]
        if node_range[0] is None:
            node_range[0] = node_attribute.min().min()
        if node_range[1] is None:
            node_range[1] = node_attribute.max().max()
    else:
        node_index = None
        initial_node_values = None
        
    if link_attribute is not None:
        link_index = link_attribute.index
        initial_link_values = link_attribute.iloc[0, :]
        if link_range[0] is None:
            link_range[0] = link_attribute.min().min()
        if link_range[1] is None:
            link_range[1] = link_attribute.max().max()
    else:
        link_index = None
        initial_link_values = None
    
    if (node_index is not None) & (link_index is not None):
        if len(node_index.symmetric_difference(link_index)) > 0:
            print('Node attribute index does not equal link attribute index')
            return
        index = node_index
    elif node_index is not None:
        index = node_index
    elif link_index is not None:
        index = link_index
    else:
        print('Node attributes or link attributes must be included')
        return
    
    if ax is None: # create a new figure
        fig = plt.figure(facecolor='w', edgecolor='k')
        ax = plt.gca()
            
    if title is not None:
        title_name = title + ', ', str(index[0])
    else:
        title_name = '0'
    
    ax = plot_network(wn, node_attribute=initial_node_values, link_attribute=initial_link_values, title=title_name,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, directed=directed, ax=ax, show_plot=False)
        
    def update(n):
        if node_attribute is not None:
            node_values = node_attribute.iloc[n, :]
        else:
            node_values = None
        
        if link_attribute is not None:
            link_values = link_attribute.iloc[n, :]
        else:
            link_values = None
            
        if title is not None:
            title_name = title + ', ' + str(index[n])
        else:
            title_name = str(n)
        
        fig.clf()  
        ax = plt.gca()
        
        ax = plot_network(wn, node_attribute=node_values, link_attribute=link_values, title=title_name,
               node_size=node_size, node_range=node_range, node_alpha=node_alpha, node_cmap=node_cmap, node_labels=node_labels,
               link_width=link_width, link_range=link_range, link_alpha=link_alpha, link_cmap=link_cmap, link_labels=link_labels,
               add_colorbar=add_colorbar, directed=directed, ax=ax, show_plot=False)
        
        return ax
    
    anim = animation.FuncAnimation(fig, update, interval=50, frames=len(index), blit=False, repeat=repeat)
    
    return anim


"""
The wntr.network package contains methods to define a water network model,
network controls, and water network model I/O.
"""
from .base import Node, Link, NodeType, LinkType, LinkStatus
from .elements import Junction, Reservoir, Tank, Pipe, Pump, Valve, Pattern, \
    TimeSeries, Demands, Curve, Source
from .model import WaterNetworkModel
from .layer import generate_valve_layer
from .options import Options
from .controls import Comparison, ControlPriority, TimeOfDayCondition, \
    SimTimeCondition, ValueCondition, TankLevelCondition, RelativeCondition, \
    OrCondition, AndCondition, ControlAction, Control, ControlChecker, \
    ControlChangeTracker, Rule
from .io import to_dict, from_dict, to_gis, from_gis, to_graph, \
    read_inpfile, write_inpfile, \
    read_json, write_json, \
    read_geojson, write_geojson, \
    read_shapefile, write_shapefile



import sys
import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
netdir = join(testdir, "..", "..", "examples", "networks")



class TestMorph(unittest.TestCase):
    def test_scale_node_coordinates(self):

        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        node = wn.get_node("123")
        coord = node.coordinates
        vertex = (8.5, 27)
        wn.get_link('10').vertices.append(vertex)

        wn2 = wntr.morph.scale_node_coordinates(wn, 100)
        node2 = wn2.get_node("123")
        coord2 = node2.coordinates
        vertex2 = wn2.get_link('10').vertices[0]

        self.assertEqual(coord[0] * 100, coord2[0])
        self.assertEqual(coord[1] * 100, coord2[1])
        self.assertEqual(vertex[0] * 100, vertex2[0])
        self.assertEqual(vertex[1] * 100, vertex2[1])

    def test_translate_node_coordinates(self):

        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        node = wn.get_node("123")
        coord = node.coordinates
        vertex = (8.5, 27)
        wn.get_link('10').vertices.append(vertex)

        wn2 = wntr.morph.translate_node_coordinates(wn, 5, 10)
        node2 = wn2.get_node("123")
        coord2 = node2.coordinates
        vertex2 = wn2.get_link('10').vertices[0]

        self.assertEqual(coord[0] + 5, coord2[0])
        self.assertEqual(coord[1] + 10, coord2[1])
        self.assertEqual(vertex[0] + 5, vertex2[0])
        self.assertEqual(vertex[1] + 10, vertex2[1])

    def test_rotate_node_coordinates(self):

        wn = wntr.network.WaterNetworkModel()
        wn.add_junction("J1", base_demand=5, elevation=100.0, coordinates=(2, 0))
        wn.add_junction("J2", base_demand=5, elevation=100.0, coordinates=(8, 0))
        wn.add_pipe("P1", "J1", "J2")
        vertex = (4, 0)
        wn.get_link('P1').vertices.append(vertex)

        wn2 = wntr.morph.rotate_node_coordinates(wn, 45)
        node2 = wn2.get_node("J1")
        coord2 = node2.coordinates
        vertex2 = wn2.get_link('P1').vertices[0]

        self.assertAlmostEqual(np.sqrt(2), coord2[0], 6)
        self.assertAlmostEqual(np.sqrt(2), coord2[1], 6)
        self.assertAlmostEqual(vertex[0] * np.sqrt(2) / 2, vertex2[0])
        self.assertAlmostEqual(vertex[0] * np.sqrt(2) / 2 , vertex2[1])

    def test_UTM_to_longlat_to_UTM(self):

        wn = wntr.network.WaterNetworkModel()
        wn.add_junction(
            "J1", base_demand=5, elevation=100.0, coordinates=(351521.07, 3886097.33)
        )  # easting, northing
        wn.add_junction(
            "J2", base_demand=5, elevation=100.0, coordinates=(351700.00, 3886097.33)
        )
        wn.add_pipe("P1", "J1", "J2")
        vertex = (351600.00, 3886097.33)
        wn.get_link('P1').vertices.append(vertex)

        wn2 = wntr.morph.convert_node_coordinates_UTM_to_longlat(wn, 13, "S")
        node2 = wn2.get_node("J1")
        coord2 = node2.coordinates
        vertex2 = wn2.get_link('P1').vertices[0]

        self.assertAlmostEqual(-106.629181, coord2[0], 6)  # longitude
        self.assertAlmostEqual(35.106766, coord2[1], 6)  # latitude
        self.assertAlmostEqual(-106.628315, vertex2[0], 6)  # longitude
        self.assertAlmostEqual(35.106778, vertex2[1], 6)  # latitude

        wn3 = wntr.morph.convert_node_coordinates_longlat_to_UTM(wn2)
        node3 = wn3.get_node("J1")
        coord3 = node3.coordinates
        vertex3 = wn3.get_link('P1').vertices[0]

        self.assertAlmostEqual(351521.07, coord3[0], 1)  # easting
        self.assertAlmostEqual(3886097.33, coord3[1], 1)  # northing
        self.assertAlmostEqual(351600.00, vertex3[0], 1)  # easting
        self.assertAlmostEqual(3886097.33, vertex3[1], 1)  # northing

    def test_convert_node_coordinates_to_longlat(self):

        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        vertex = (8.5, 27)
        wn.get_link('10').vertices.append(vertex)

        longlat_map = {"Lake": (-106.6587, 35.0623), "219": (-106.5248, 35.191)}
        wn2 = wntr.morph.convert_node_coordinates_to_longlat(wn, longlat_map)
        for node_name in longlat_map.keys():
            node = wn2.get_node(node_name)
            coord = node.coordinates
            self.assertAlmostEqual(longlat_map[node_name][0], coord[0], 4)
            self.assertAlmostEqual(longlat_map[node_name][1], coord[1], 4)
        vertex2 = wn2.get_link('10').vertices[0]
        self.assertAlmostEqual(-106.6555, vertex2[0], 4)
        self.assertAlmostEqual(35.0638, vertex2[1], 4)

        # opposite rotation
        longlat_map = {"Lake": (-106.6851, 35.1344), "219": (-106.5073, 35.0713)}
        wn2 = wntr.morph.convert_node_coordinates_to_longlat(wn, longlat_map)
        for node_name in longlat_map.keys():
            node = wn2.get_node(node_name)
            coord = node.coordinates
            self.assertAlmostEqual(longlat_map[node_name][0], coord[0], 4)
            self.assertAlmostEqual(longlat_map[node_name][1], coord[1], 4)
        vertex2 = wn2.get_link('10').vertices[0]
        self.assertAlmostEqual(-106.6826, vertex2[0], 4)
        self.assertAlmostEqual(35.1325, vertex2[1], 4)

    def test_split_pipe(self):

        inp_file = join(datadir, "leaks.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn = wntr.morph.split_pipe(wn, "pipe1", "pipe1__B", "leak1")

        pipe = wn.get_link("pipe1")
        pipeB = wn.get_link("pipe1__B")

        self.assertEqual(True, "leak1" in [name for name, n in wn.nodes()])
        self.assertEqual(
            True, "leak1" in [name for name, n in wn.nodes(wntr.network.Junction)]
        )
        self.assertEqual(True, "pipe1" in [name for name, l in wn.links()])
        self.assertEqual(
            True, "pipe1" in [name for name, l in wn.links(wntr.network.Pipe)]
        )
        self.assertEqual(True, "pipe1__B" in [name for name, l in wn.links()])
        self.assertEqual(
            True, "pipe1__B" in [name for name, l in wn.links(wntr.network.Pipe)]
        )
        self.assertEqual(pipe.end_node_name, "leak1")
        self.assertEqual(pipeB.start_node_name, "leak1")
        self.assertEqual(pipe.diameter, pipeB.diameter)
        self.assertEqual(pipe.roughness, pipeB.roughness)
        self.assertEqual(pipe.minor_loss, pipeB.minor_loss)
        self.assertEqual(pipe.initial_status, pipeB.initial_status)
    
    def test_break_pipe(self):

        inp_file = join(datadir, "leaks.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn = wntr.morph.break_pipe(wn, "pipe1", "pipe1__B", "leak1", "leak2")

        pipe = wn.get_link("pipe1")
        pipeB = wn.get_link("pipe1__B")

        self.assertEqual(True, "leak1" in [name for name, n in wn.nodes()])
        self.assertEqual(
            True, "leak1" in [name for name, n in wn.nodes(wntr.network.Junction)]
        )
        self.assertEqual(True, "leak2" in [name for name, n in wn.nodes()])
        self.assertEqual(
            True, "leak2" in [name for name, n in wn.nodes(wntr.network.Junction)]
        )
        self.assertEqual(True, "pipe1" in [name for name, l in wn.links()])
        self.assertEqual(
            True, "pipe1" in [name for name, l in wn.links(wntr.network.Pipe)]
        )
        self.assertEqual(True, "pipe1__B" in [name for name, l in wn.links()])
        self.assertEqual(
            True, "pipe1__B" in [name for name, l in wn.links(wntr.network.Pipe)]
        )
        self.assertEqual(pipe.end_node_name, "leak1")
        self.assertEqual(pipeB.start_node_name, "leak2")
        self.assertEqual(pipe.diameter, pipeB.diameter)
        self.assertEqual(pipe.roughness, pipeB.roughness)
        self.assertEqual(pipe.minor_loss, pipeB.minor_loss)
        self.assertEqual(pipe.initial_status, pipeB.initial_status)

    def test_split_break_pipe_vertices(self):

        inp_file = join(datadir, "io.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        
        wn2 = wntr.morph.split_pipe(wn, "p1", "p1__new", "new_junc")
        pipe = wn2.get_link("p1")
        pipeB = wn2.get_link("p1__new")
        self.assertEqual(len(pipe.vertices), 0)
        self.assertEqual(len(pipeB.vertices), 2)
        
        wn2 = wntr.morph.split_pipe(wn, "p1", "p1__new", "new_junc", split_at_point=0.7)
        pipe = wn2.get_link("p1")
        pipeB = wn2.get_link("p1__new")
        self.assertEqual(pipe.vertices, [(15.0, 5.0)])
        self.assertEqual(pipeB.vertices, [(20.0, 5.0)])
        
        wn2 = wntr.morph.split_pipe(wn, "p1", "p1__new", "new_junc", split_at_point=0.9)
        pipe = wn2.get_link("p1")
        pipeB = wn2.get_link("p1__new")
        self.assertEqual(len(pipe.vertices), 2)
        self.assertEqual(len(pipeB.vertices), 0)

    def test_reverse_pipes(self):
        inp_file = join(datadir, "io.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn2 = wntr.morph.link.reverse_link(wn, "p1")
        pipe2 = wn2.get_link("p1")

        # test start and end nodes
        self.assertEqual(pipe2.start_node, wn2.get_node('j1'))
        self.assertEqual(pipe2.end_node, wn2.get_node('t1'))

        # test vertices
        self.assertEqual(pipe2.vertices, [(20.0, 5.0), (15.0, 5.0)])

    def test_skeletonize(self):

        inp_file = join(datadir, "skeletonize.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        expected_total_demand = 12.1 / 264.172 / 60  # 12.1 GPM converted to m3/s

        expected_nums = pd.DataFrame(
            index=[0, 4, 8, 12, 24, 36], columns=["num_nodes", "num_links"]
        )
        expected_nums.loc[0, :] = [wn.num_nodes, wn.num_links]
        expected_nums.loc[4, :] = [wn.num_nodes - 5, wn.num_links - 5]
        expected_nums.loc[8, :] = [wn.num_nodes - 15, wn.num_links - 18]
        expected_nums.loc[12, :] = [wn.num_nodes - 21, wn.num_links - 26]
        expected_nums.loc[24, :] = [wn.num_nodes - 25, wn.num_links - 30]
        expected_nums.loc[36, :] = [wn.num_nodes - 29, wn.num_links - 34]

        for i in [0, 4, 8, 12, 24, 36]:
            skel_wn, skel_map = wntr.morph.skeletonize(
                wn, float(i) * 0.0254, return_map=True, use_epanet=False
            )

            demand = wntr.metrics.expected_demand(skel_wn)
            total_demand = demand.loc[0, :].sum()

            # Write skel_wn to an inp file, read it back in, then extract the demands
            skel_inp_file = "temp.inp"
            wntr.network.write_inpfile(skel_wn, skel_inp_file, "GPM")
            skel_wn_io = wntr.network.WaterNetworkModel(skel_inp_file)
            demand_io = wntr.metrics.expected_demand(skel_wn_io)
            total_demand_io = demand_io.loc[0, :].sum()

            # pipes = wn.query_link_attribute('diameter', np.less_equal, i*0.0254)
            # wntr.graphics.plot_network(wn, link_attribute = list(pipes.keys()), title=str(i))
            # wntr.graphics.plot_network(skel_wn, link_attribute='diameter', link_width=2, node_size=15, title=str(i))

            self.assertAlmostEqual(total_demand.sum(), expected_total_demand, 6)
            self.assertAlmostEqual(total_demand_io.sum(), expected_total_demand, 6)

            self.assertEqual(skel_wn.num_nodes, expected_nums.loc[i, "num_nodes"])
            self.assertEqual(skel_wn.num_links, expected_nums.loc[i, "num_links"])

            if i == 0:
                expected_map = {}  # 1:1 map
                for name in wn.node_name_list:
                    expected_map[name] = [name]
                self.assertEqual(dict(expected_map, **skel_map), skel_map)

            if i == 4:
                expected_map_subset = {}
                expected_map_subset["15"] = ["15", "14", "16"]
                expected_map_subset["30"] = ["30", "32"]
                expected_map_subset["56"] = ["56", "57"]
                expected_map_subset["59"] = ["59", "64"]
                expected_map_subset["14"] = []
                expected_map_subset["16"] = []
                expected_map_subset["32"] = []
                expected_map_subset["57"] = []
                expected_map_subset["64"] = []
                self.assertEqual(dict(expected_map_subset, **skel_map), skel_map)

    def test_skeletonize_with_controls(self):

        inp_file = join(datadir, "skeletonize.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        
        # Run skeletonization without excluding junctions or pipes
        skel_wn = wntr.morph.skeletonize(wn, 12.0 * 0.0254, use_epanet=False)
        # Junction 13 and Pipe 60 are not in the skeletonized model
        assert "13" not in skel_wn.junction_name_list
        assert "60" not in skel_wn.pipe_name_list
        
        # add control to a link
        action = wntr.network.ControlAction(
            wn.get_link("60"), "status", wntr.network.LinkStatus.Closed
        )
        condition = wntr.network.SimTimeCondition(wn, "==", 0)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("close_valve", control)

        # add control to a node
        action = wntr.network.ControlAction(wn.get_node("13"), "elevation", 1)
        condition = wntr.network.SimTimeCondition(wn, "==", 0)
        control = wntr.network.Control(condition=condition, then_action=action)
        wn.add_control("raise_node", control)

        # Rerun skeletonize
        skel_wn = wntr.morph.skeletonize(wn, 12.0 * 0.0254, use_epanet=False)
        assert "13" in skel_wn.junction_name_list
        assert "60" in skel_wn.pipe_name_list
        self.assertEqual(skel_wn.num_nodes, wn.num_nodes - 17)
        self.assertEqual(skel_wn.num_links, wn.num_links - 22)

    def test_skeletonize_with_excluding_nodes_and_pipes(self):

        inp_file = join(datadir, "skeletonize.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        
        # Run skeletonization without excluding junctions or pipes
        skel_wn = wntr.morph.skeletonize(wn, 12.0 * 0.0254, use_epanet=False)
        # Junction 13 and Pipe 60 are not in the skeletonized model
        assert "13" not in skel_wn.junction_name_list
        assert "60" not in skel_wn.pipe_name_list
        
        # Run skeletonization excluding Junction 13 and Pipe 60
        skel_wn = wntr.morph.skeletonize(wn, 12.0 * 0.0254, use_epanet=False, 
                                         junctions_to_exclude=["13"], 
                                         pipes_to_exclude=["60"])
        # Junction 13 and Pipe 60 are in the skeletonized model
        assert "13" in skel_wn.junction_name_list
        assert "60" in skel_wn.pipe_name_list
        self.assertEqual(skel_wn.num_nodes, wn.num_nodes - 17)
        self.assertEqual(skel_wn.num_links, wn.num_links - 22)

        # Change diameter of link 60 one link connected to Junction 13 to be 
        # greater than 12, should get some results as above
        # Note, link 11 is connected to Junction 13
        link = wn.get_link("60")
        link.diameter = 16 * 0.0254
        link_connected_to_13 = wn.get_links_for_node('13')[0]
        link = wn.get_link(link_connected_to_13)
        link.diameter = 16 * 0.0254

        skel_wn = wntr.morph.skeletonize(wn, 12.0 * 0.0254, use_epanet=False)
        assert "13" in skel_wn.junction_name_list
        assert "60" in skel_wn.pipe_name_list
        self.assertEqual(skel_wn.num_nodes, wn.num_nodes - 17)
        self.assertEqual(skel_wn.num_links, wn.num_links - 22)

    def test_series_merge_properties(self):

        wn = wntr.network.WaterNetworkModel()

        wn.add_junction("J1", base_demand=5, elevation=100.0, coordinates=(0, 0))
        wn.add_junction("J2", base_demand=8, elevation=50.0, coordinates=(1, 0))
        wn.add_junction("J3", base_demand=5, elevation=25.0, coordinates=(2, 0))
        wn.add_pipe(
            "P12",
            "J1",
            "J2",
            length=350,
            diameter=8,
            roughness=120,
            minor_loss=0.1,
            initial_status="OPEN",
        )
        wn.add_pipe(
            "P23",
            "J2",
            "J3",
            length=250,
            diameter=6,
            roughness=80,
            minor_loss=0.0,
            initial_status="OPEN",
        )

        # Add a source
        wn.add_reservoir("R", base_head=125, coordinates=(0, 2))
        wn.add_pipe(
            "PR",
            "R",
            "J1",
            length=100,
            diameter=12,
            roughness=100,
            minor_loss=0.0,
            initial_status="OPEN",
        )

        wn.options.time.duration = 0

        skel_wn = wntr.morph.skeletonize(
            wn,
            8,
            branch_trim=False,
            series_pipe_merge=True,
            parallel_pipe_merge=False,
            max_cycles=1,
            use_epanet=False,
        )

        link = skel_wn.get_link("P12")  # pipe P12 is the dominant pipe

        self.assertEqual(link.length, 600)
        self.assertEqual(link.diameter, 8)
        self.assertAlmostEqual(link.roughness, 55, 0)
        self.assertEqual(link.minor_loss, 0.1)
        self.assertEqual(link.status, 1)  # open

    def test_parallel_merge_properties(self):

        wn = wntr.network.WaterNetworkModel()

        wn.add_junction("J1", base_demand=5, elevation=100.0, coordinates=(0, 0))
        wn.add_junction("J2", base_demand=8, elevation=50.0, coordinates=(1, 0))
        wn.add_pipe(
            "P12a",
            "J1",
            "J2",
            length=280,
            diameter=250,
            roughness=120,
            minor_loss=0.1,
            initial_status="OPEN",
        )
        wn.add_pipe(
            "P12b",
            "J1",
            "J2",
            length=220,
            diameter=300,
            roughness=100,
            minor_loss=0,
            initial_status="OPEN",
        )
        # Add a source
        wn.add_reservoir("R", base_head=125, coordinates=(0, 2))
        wn.add_pipe(
            "PR",
            "R",
            "J1",
            length=100,
            diameter=450,
            roughness=100,
            minor_loss=0.0,
            initial_status="OPEN",
        )

        wn.options.time.duration = 0

        skel_wn = wntr.morph.skeletonize(
            wn,
            300,
            branch_trim=False,
            series_pipe_merge=False,
            parallel_pipe_merge=True,
            max_cycles=1,
            use_epanet=False,
        )

        link = skel_wn.get_link("P12b")  # pipe P12b is the dominant pipe

        self.assertEqual(link.length, 220)
        self.assertEqual(link.diameter, 300)
        self.assertAlmostEqual(link.roughness, 165, 0)
        self.assertEqual(link.minor_loss, 0)
        self.assertEqual(link.status, 1)  # open

    def test_skeletonize_Net3(self):

        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        skel_wn = wntr.morph.skeletonize(wn, 36.0 * 0.0254, use_epanet=False)

        self.assertEqual(wn.num_junctions, 92)
        self.assertEqual(skel_wn.num_junctions, 45)

        sim = wntr.sim.WNTRSimulator(wn)
        results_original = sim.run_sim()

        sim = wntr.sim.WNTRSimulator(skel_wn)
        results_skel = sim.run_sim()

        skel_junctions = skel_wn.junction_name_list

        pressure_orig = results_original.node["pressure"].loc[:, skel_junctions]
        pressure_skel = results_skel.node["pressure"].loc[:, skel_junctions]
        pressure_diff = abs(pressure_orig - pressure_skel)
        pressure_diff.index = pressure_diff.index / 3600

        m50 = pressure_diff.quantile(0.50, axis=1)

        """
        import matplotlib.pylab as plt
        wntr.graphics.plot_network(wn, title='Original')
        wntr.graphics.plot_network(skel_wn, title='Skeletonized')
        plt.figure()
        m50.plot()
        print(m50.mean())
        """
        self.assertLess(m50.max(), 1.5)
        self.assertLess(m50.mean(), 0.15)


if __name__ == "__main__":
    unittest.main()



"""
The wntr.graphics package contains graphic functions
"""
from wntr.graphics.network import plot_network, plot_interactive_network, plot_leaflet_network, network_animation
from wntr.graphics.layer import plot_valve_layer
from wntr.graphics.curve import plot_fragility_curve, plot_pump_curve, plot_tank_volume_curve
from wntr.graphics.color import custom_colormap, random_colormap




import os
import sys
import unittest
import pytest
from os import listdir
from os.path import abspath, dirname, isfile, join
import nbformat
import pandas as pd
import geopandas as gpd
from pandas.testing import assert_frame_equal
from nbconvert.preprocessors import ExecutePreprocessor

kernel_name = 'python%d' % sys.version_info[0]
testdir = dirname(abspath(str(__file__)))
examplesdir = join(testdir, '..', '..', 'examples')

class TestDemos(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr
        
        self.nb_files = {
            ##TODO add file comparison tests for the additional tutorials
            #'basics_tutorial': ['Net3_analysis_pipes.geojson'],
            'earthquake_tutorial': ['earthquake_people_impacted.csv',
                                    'earthquake_people_impacted_wrepair.csv'],
            'fire_flow_tutorial': ['fire_flow_junctions_impacted.csv',
                                   'fire_flow_people_impacted.csv'],
            'getting_started_tutorial': ['pressure.csv'],
            #'landslide_tutorial': ['ky10_landslide_analysis_junctions.geojson'],
            'model_development_tutorial': ['wn1.inp', 
                                           'wn2.inp'],
            'multispecies_tutorial': ['chlorine_residual_1.3.csv'],
            'pipe_break_tutorial': ['pipe_break_junctions_impacted.csv',
                                    'pipe_break_people_impacted.csv'],
            'pipe_segments_tutorial': ['segment_break_junctions_impacted.csv',
                                       'segment_break_people_impacted.csv'],
            'salt_water_intrusion_tutorial': ['salt_water_baseline_quality.csv',
                                              'salt_water_response_quality.csv']
            }

    @classmethod
    def tearDownClass(self):
        pass

    def results_test(self, file1):
        # Check if jupyter notebook results (in file1) match expected results (in file2)
        file_extension = file1.split('.')[-1]
        assert file_extension in ['csv', 'geojson', 'inp'], file1
        
        file2 = join(testdir, 'data_for_testing', file1)

        if file_extension in ['csv', 'geojson']:
            if file_extension == 'csv':
                results1 = pd.read_csv(file1, index_col=0)
                results2 = pd.read_csv(file2, index_col=0)
            if file_extension == 'geojson':
                results1 = gpd.read_file(file1, index_col=0)
                results2 = gpd.read_file(file2, index_col=0)
            
            for axis in [0,1]:
                results1.sort_index(axis=axis, inplace=True)
                results2.sort_index(axis=axis, inplace=True)

        elif file_extension == 'inp':
            wn1 = self.wntr.network.WaterNetworkModel(file1)
            wn2 = self.wntr.network.WaterNetworkModel(file2)
            
            results1 = pd.DataFrame(wn1.describe(level=2))
            results2 = pd.DataFrame(wn2.describe(level=2))
        
        try:
            assert_frame_equal(results1, results2, rtol=1e-2, atol=1e-4)
            return True
        except:
            return False
        

    @pytest.mark.time_consuming
    def test_that_demos_run(self):
        cwd = os.getcwd()
        os.chdir(examplesdir)
        example_files = [
            f
            for f in listdir(examplesdir)
            if isfile(join(examplesdir, f))
            and f.endswith(".ipynb")
            and not f.startswith("test")
        ]
        
        flag = 0
        failed_examples = []
        for f in example_files:           
            print(f)
            nb_name, file_extension = os.path.splitext(os.path.basename(f))
            
            with open(f) as file:
                nb = nbformat.read(file, as_version=4) 
                
            nb.metadata.get('kernelspec', {})['name'] = kernel_name            
            proc = ExecutePreprocessor(timeout=600, kernel_name=kernel_name)
            proc.allow_errors = True
            proc.preprocess(nb)
            
            # Test to make sure the demo runs
            runtime_errors = []
            for cell in nb.cells:
                if 'outputs' in cell:
                    for output in cell['outputs']:
                        if output.output_type == 'error':
                            runtime_errors.append(output)
            print('  runtime errors', len(runtime_errors))
            if len(runtime_errors) > 0:
                failed_examples.append(f)
                flag = 1
                continue
            
            # Test to make sure results are the same (based on a comparison of
            # results saved to file)
            file_errors = 0
            if nb_name in self.nb_files.keys():
                for file in self.nb_files[nb_name]:
                    if not self.results_test(file):
                        file_errors = file_errors + 1
                print('  file comparison errors', file_errors)
                if file_errors > 0:
                    failed_examples.append(f)
                    flag = 1
                    
        os.chdir(cwd)
        if len(failed_examples) > 0:
            print("failed demos: {0}".format(failed_examples))
        self.assertEqual(flag, 0)

if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import abspath, dirname, join
from unittest import SkipTest

import networkx as nx
import numpy as np
import wntr

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
netdir = join(testdir, "..", "..", "examples", "networks")


class TestNetworkGraphs(unittest.TestCase):
    def test_weight_graph(self):
        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        node_weight = wn.query_node_attribute("elevation")
        link_weight = wn.query_link_attribute("length")

        G = wn.to_graph(node_weight, link_weight)

        self.assertEqual(G.nodes["111"]["weight"], 10 * 0.3048)
        self.assertEqual(G["159"]["161"]["177"]["weight"], 2000 * 0.3048)

    def test_weighted_graph_modify_direction(self):
        inp_file = join(netdir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file) 
        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim()
        flowrate = results.link['flowrate'].iloc[-1,:]
        G = wn.to_graph(link_weight=flowrate, modify_direction=True)
        
        # Positive flow, flowrate == graph weight
        name = '173'
        pipe = wn.get_link(name)

        self.assertEqual(flowrate[name],
                         G.edges[pipe.start_node_name, pipe.end_node_name, name]['weight'])
        
        # Negative flow, -flowrate == graph weight
        name = '109'
        pipe = wn.get_link(name)
        
        self.assertEqual(-flowrate[name],
                         G.edges[pipe.end_node_name, pipe.start_node_name, name]['weight'])

    def test_terminal_nodes(self):
        inp_file = join(netdir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        G = wn.to_graph()

        terminal_nodes = wntr.metrics.terminal_nodes(G)
        expected = set(["2", "9"])
        self.assertSetEqual(set(terminal_nodes), expected)

    def test_bridges(self):
        inp_file = join(netdir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        G = wn.to_graph()

        bridges = wntr.metrics.bridges(G)
        expected = set(["9", "10", "110"])
        self.assertSetEqual(set(bridges), expected)

    def test_diameter(self):
        inp_file = join(datadir, "Anytown.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        for pump in wn.pump_name_list[:-1]:  # remove 2 of the 3 pumps
            wn.remove_link(pump)
        G = wn.to_graph()
        udG = G.to_undirected()
        val = nx.diameter(udG)
        excepted = 7  # Davide Soldi et al. (2015) Procedia Engineering
        self.assertEqual(val, excepted)

    def test_central_point_dominance(self):
        inp_file = join(datadir, "Anytown.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        for pump in wn.pump_name_list[:-1]:  # remove 2 of the 3 pumps
            wn.remove_link(pump)
        G = wn.to_graph()

        val = wntr.metrics.central_point_dominance(G)
        expected = 0.23  # Davide Soldi et al. (2015) Procedia Engineering
        error = abs(expected - val)
        self.assertLess(error, 0.01)

    def test_spectral_gap(self):
        inp_file = join(datadir, "Anytown.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        for pump in wn.pump_name_list[:-1]:  # remove 2 of the 3 pumps
            wn.remove_link(pump)
        G = wn.to_graph()

        val = wntr.metrics.spectral_gap(G)
        expected = 1.5149  # Davide Soldi et al. (2015) Procedia Engineering
        error = abs(expected - val)
        self.assertLess(error, 0.01)

    def test_algebraic_connectivity(self):
        inp_file = join(datadir, "Anytown.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        for pump in wn.pump_name_list[:-1]:  # remove 2 of the 3 pumps
            wn.remove_link(pump)
        G = wn.to_graph()

        val = wntr.metrics.algebraic_connectivity(G)
        expected = 0.1708  # Davide Soldi et al. (2015) Procedia Engineering
        error = abs(expected - val)
        raise SkipTest
        self.assertLess(error, 0.01)

    def test_crit_ratio_defrag(self):
        inp_file = join(datadir, "Anytown.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        for pump in wn.pump_name_list[:-1]:  # remove 2 of the 3 pumps
            wn.remove_link(pump)
        G = wn.to_graph()

        val = wntr.metrics.critical_ratio_defrag(G)
        expected = 0.63  # Pandit et al. (2012) Critical Infrastucture Symposium
        error = abs(expected - val)
        raise SkipTest
        self.assertLess(error, 0.01)

    def test_Net1_MultiDiGraph(self):
        inp_file = join(netdir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        G = wn.to_graph()

        node = {
            "11": {"pos": (30.0, 70.0), "type": "Junction"},
            "10": {"pos": (20.0, 70.0), "type": "Junction"},
            "13": {"pos": (70.0, 70.0), "type": "Junction"},
            "12": {"pos": (50.0, 70.0), "type": "Junction"},
            "21": {"pos": (30.0, 40.0), "type": "Junction"},
            "22": {"pos": (50.0, 40.0), "type": "Junction"},
            "23": {"pos": (70.0, 40.0), "type": "Junction"},
            "32": {"pos": (50.0, 10.0), "type": "Junction"},
            "31": {"pos": (30.0, 10.0), "type": "Junction"},
            "2": {"pos": (50.0, 90.0), "type": "Tank"},
            "9": {"pos": (10.0, 70.0), "type": "Reservoir"},
        }

        edge = {
            "11": {"12": {"11": {"type": "Pipe"}}, "21": {"111": {"type": "Pipe"}}},
            "10": {"11": {"10": {"type": "Pipe"}}},
            "13": {"23": {"113": {"type": "Pipe"}}},
            "12": {"13": {"12": {"type": "Pipe"}}, "22": {"112": {"type": "Pipe"}}},
            "21": {"31": {"121": {"type": "Pipe"}}, "22": {"21": {"type": "Pipe"}}},
            "22": {"32": {"122": {"type": "Pipe"}}, "23": {"22": {"type": "Pipe"}}},
            "23": {},
            "32": {},
            "31": {"32": {"31": {"type": "Pipe"}}},
            "2": {"12": {"110": {"type": "Pipe"}}},
            "9": {"10": {"9": {"type": "Pump"}}},
        }

        self.assertEqual(dict(node, **G.nodes), node)
        # assert_dict_contains_subset(node, G.nodes)
        self.assertEqual(dict(edge, **G.adj), edge)
        # assert_dict_contains_subset(edge, G.adj)


if __name__ == "__main__":
    unittest.main()



"""WNTR Simulator hydraulics model."""

import pandas as pd
import numpy as np
import scipy.sparse as sparse
import math
import warnings
import logging
from wntr.network.model import WaterNetworkModel
from wntr.network.base import NodeType, LinkType, LinkStatus
from wntr.network.elements import Junction, Tank, Reservoir, Pipe, Pump, HeadPump, PowerPump, PRValve, PSValve, FCValve, \
    TCValve, GPValve, PBValve
from collections import OrderedDict
from wntr.utils.ordered_set import OrderedSet
from wntr.sim import aml
from wntr.sim.models import constants, var, param, constraint
from wntr.sim.models.utils import ModelUpdater

logger = logging.getLogger(__name__)


def create_hydraulic_model(wn, HW_approx='default'):
    """
    Parameters
    ----------
    wn: WaterNetworkModel
    mode: str
    HW_approx: str
        Specifies which Hazen-Williams headloss approximation to use. Options are 'default' and 'piecewise'. Please
        see the WNTR documentation on hydraulics for details.

    Returns
    -------
    m: wntr.aml.Model
    model_updater: wntr.models.utils.ModelUpdater
    """
    m = aml.Model()
    model_updater = ModelUpdater()

    # Global constants
    constants.hazen_williams_constants(m)
    constants.head_pump_constants(m)
    constants.leak_constants(m)
    constants.pdd_constants(m)

    param.source_head_param(m, wn)
    param.expected_demand_param(m, wn)
    mode = wn.options.hydraulic.demand_model
    if mode in ['DD', 'DDA']:
        pass
    elif mode in ['PDD', 'PDA']:
        param.pmin_param.build(m, wn, model_updater)
        param.pnom_param.build(m, wn, model_updater)
        param.pdd_poly_coeffs_param.build(m, wn, model_updater)
    param.leak_coeff_param.build(m, wn, model_updater)
    param.leak_area_param.build(m, wn, model_updater)
    param.leak_poly_coeffs_param.build(m, wn, model_updater)
    param.elevation_param.build(m, wn, model_updater)
    if wn.options.hydraulic.headloss == 'C-M':
        raise NotImplementedError('C-M headloss is not currently supported in the WNTRSimulator')
    if wn.options.hydraulic.headloss == 'D-W':
        raise NotImplementedError('D-W headloss is not currently supported in the WNTRSimulator')
    param.hw_resistance_param.build(m, wn, model_updater)
    param.minor_loss_param.build(m, wn, model_updater)
    param.tcv_resistance_param.build(m, wn, model_updater)
    param.pump_power_param.build(m, wn, model_updater)
    param.valve_setting_param.build(m, wn, model_updater)

    if mode in ['DD','DDA']:
        pass
    elif mode in ['PDD','PDA']:
        var.demand_var(m, wn)
    var.flow_var(m, wn)
    var.head_var(m, wn)
    var.leak_rate_var(m, wn)

    if mode in ['DD','DDA']:
        constraint.mass_balance_constraint.build(m, wn, model_updater)
    elif mode in ['PDD','PDA']:
        constraint.pdd_mass_balance_constraint.build(m, wn, model_updater)
        constraint.pdd_constraint.build(m, wn, model_updater)
    else:
        raise ValueError('mode not recognized: ' + str(mode))
    if HW_approx == 'default':
        constraint.approx_hazen_williams_headloss_constraint.build(m, wn, model_updater)
    elif HW_approx == 'piecewise':
        constraint.piecewise_hazen_williams_headloss_constraint.build(m, wn, model_updater)
    else:
        raise ValueError('Unexpected value for HW_approx: ' + str(HW_approx))
    constraint.head_pump_headloss_constraint.build(m, wn, model_updater)
    constraint.power_pump_headloss_constraint.build(m, wn, model_updater)
    constraint.prv_headloss_constraint.build(m, wn, model_updater)
    constraint.psv_headloss_constraint.build(m, wn, model_updater)
    constraint.tcv_headloss_constraint.build(m, wn, model_updater)
    constraint.fcv_headloss_constraint.build(m, wn, model_updater)
    if len(wn.pbv_name_list) > 0:
        raise NotImplementedError('PBV valves are not currently supported in the WNTRSimulator')
    if len(wn.gpv_name_list) > 0:
        raise NotImplementedError('GPV valves are not currently supported in the WNTRSimulator')
    constraint.leak_constraint.build(m, wn, model_updater)

    # TODO: Document that changing a curve with controls does not do anything; you have to change the pump_curve_name attribute on the pump

    return m, model_updater


def update_model_for_controls(m, wn, model_updater, change_tracker):
    """

    Parameters
    ----------
    m: wntr.sim.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    model_updater: wntr.sim.models.utils.ModelUpdater
    change_tracker: wntr.network.controls.ControlChangeTracker
    """
    for obj, attr in change_tracker.get_changes(ref_point='model'):
        model_updater.update(m, wn, obj, attr)
    change_tracker.reset_reference_point(key='model')


def update_model_for_isolated_junctions_and_links(m, wn, model_updater, prev_isolated_junctions, prev_isolated_links,
                                                  isolated_junctions, isolated_links):
    """

    Parameters
    ----------
    m: wntr.sim.aml.aml.Model
    wn: wntr.network.model.WaterNetworkModel
    model_updater: wntr.sim.models.utils.ModelUpdater
    prev_isolated_junctions: set
    prev_isolated_links: set
    isolated_junctions: set
    isolated_links: set
    """
    j1 = prev_isolated_junctions - isolated_junctions
    j2 = isolated_junctions - prev_isolated_junctions
    j = j1.union(j2)
    for _j in j:
        junction = wn.get_node(_j)
        model_updater.update(m, wn, junction, '_is_isolated')

    l1 = prev_isolated_links - isolated_links
    l2 = isolated_links - prev_isolated_links
    l = l1.union(l2)
    for _l in l:
        link = wn.get_link(_l)
        model_updater.update(m, wn, link, '_is_isolated')


def update_network_previous_values(wn):
    """
    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel
    """
    wn._prev_sim_time = wn.sim_time
    for link_name, link in wn.valves():
        link._prev_setting = link.setting
    for tank_name, tank in wn.tanks():
        tank._prev_head = tank.head

def update_tank_heads(wn):
    """
    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel
    """
    dt = wn.sim_time - wn._prev_sim_time   

    for tank_name, tank in wn.tanks():
        q_net = tank.demand
        dV = q_net * dt
        
        if tank.vol_curve is None:    
            delta_h = 4.0 * dV / (math.pi * tank.diameter ** 2)
        else:
            vcurve = np.array(tank.vol_curve.points)
            level_x = vcurve[:,0]
            volume_y = vcurve[:,1]
            
            # I had to include this because the _prev_head is the reference
            # point needed if the tank.head (and tank.level) have already
            # been updated. This isn't a problem for cases with no volume curve.
            if tank.head == tank._prev_head:
                cur_level = tank.level
            else:
                cur_level = tank._prev_head - (tank.head - tank.level)
                            
            V0 = np.interp(cur_level,level_x,volume_y)
            V1 = V0 + dV
            level_new = np.interp(V1,volume_y,level_x)
            delta_h = level_new - cur_level
                
        tank._head = tank._prev_head + delta_h
            


def initialize_results_dict(wn):
    """
    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel

    Returns
    -------
    res: dict
    """
    node_res = OrderedDict()
    link_res = OrderedDict()

    node_res['head'] = OrderedDict((name, list()) for name, obj in wn.nodes())
    node_res['demand'] = OrderedDict((name, list()) for name, obj in wn.nodes())
    node_res['pressure'] = OrderedDict((name, list()) for name, obj in wn.nodes())
    node_res['leak_demand'] = OrderedDict((name, list()) for name, obj in wn.nodes())

    link_res['flowrate'] = OrderedDict((name, list()) for name, obj in wn.links())
    link_res['velocity'] = OrderedDict((name, list()) for name, obj in wn.links())
    link_res['status'] = OrderedDict((name, list()) for name, obj in wn.links())
    link_res['setting'] = OrderedDict((name, list()) for name, obj in wn.links())

    return node_res, link_res


def save_results(wn, node_res, link_res):
    """
    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel
    node_res: collections.OrderedDict
    link_res: collections.OrderedDict
    """
    for name, node in wn.junctions():
        node_res['head'][name].append(node.head)
        node_res['demand'][name].append(node.demand)
        if node._is_isolated:
            node_res['pressure'][name].append(0.0)
        else:
            node_res['pressure'][name].append(node.head - node.elevation)
        node_res['leak_demand'][name].append(node.leak_demand)

    for name, node in wn.tanks():
        node_res['head'][name].append(node.head)
        node_res['demand'][name].append(node.demand)
        node_res['pressure'][name].append(node.head - node.elevation)
        node_res['leak_demand'][name].append(node.leak_demand)

    for name, node in wn.reservoirs():
        node_res['head'][name].append(node.head)
        node_res['demand'][name].append(node.demand)
        node_res['pressure'][name].append(0.0)
        node_res['leak_demand'][name].append(0.0)

    for name, link in wn.pipes():
        link_res['flowrate'][name].append(link.flow)
        link_res['velocity'][name].append(abs(link.flow)*4.0 / (math.pi*link.diameter**2))
        link_res['status'][name].append(link.status)
        link_res['setting'][name].append(link.roughness)

    for name, link in wn.head_pumps():
        link_res['flowrate'][name].append(link.flow)
        link_res['velocity'][name].append(0)
        link_res['status'][name].append(link.status)
        link_res['setting'][name].append(1)
        
        A, B, C = link.get_head_curve_coefficients()
        if link.flow > (A/B)**(1.0/C):
            start_node_name = link.start_node_name
            end_node_name = link.end_node_name
            start_node = wn.get_node(start_node_name)
            end_node = wn.get_node(end_node_name)
            start_head = start_node.head
            end_head = end_node.head
            warnings.warn('Pump ' + name + ' has exceeded its maximum flow.')
            logger.warning(
                'Pump {0} has exceeded its maximum flow. Pump head: {1}; Pump flow: {2}; Max pump flow: {3}'.format(
                    name, end_head - start_head, link.flow, (A/B)**(1.0/C)))

    for name, link in wn.power_pumps():
        link_res['flowrate'][name].append(link.flow)
        link_res['velocity'][name].append(0)
        link_res['status'][name].append(link.status)
        link_res['setting'][name].append(1) # power pumps have no speed

    for name, link in wn.valves():
        link_res['flowrate'][name].append(link.flow)
        link_res['velocity'][name].append(abs(link.flow)*4.0 / (math.pi*link.diameter**2))
        link_res['status'][name].append(link.status)
        link_res['setting'][name].append(link.setting)


def get_results(wn, results, node_res, link_res):
    """
    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel
    results: wntr.sim.results.SimulationResults
    node_res: collections.OrderedDict
    link_res: collections.OrderedDict
    """
    ntimes = len(results.time)
    nnodes = wn.num_nodes
    nlinks = wn.num_links
    node_names = wn.junction_name_list + wn.tank_name_list + wn.reservoir_name_list
    link_names = wn.pipe_name_list + wn.head_pump_name_list + wn.power_pump_name_list + wn.valve_name_list

    for key, value in node_res.items():
        node_res[key] = pd.DataFrame(data=np.array([node_res[key][name] for name in node_names]).transpose(), index=results.time,
                                     columns=node_names)
    results.node = node_res

    for key, value in link_res.items():
        link_res[key] = pd.DataFrame(data=np.array([link_res[key][name] for name in link_names]).transpose(), index=results.time,
                                            columns=link_names)
    results.link = link_res
    
    # Add headloss to results.link -- removed for now, this is slow
    #headloss = pd.DataFrame(data=None, index=results.time, columns=link_names)
    # for name, link in wn.links():
    #     start_node = link.start_node_name
    #     end_node = link.end_node_name
    #     start_head = results.node['head'].loc[:,start_node]
    #     end_head = results.node['head'].loc[:,end_node]
    #     if isinstance(link, Pipe):
    #         # Unit headloss for pipes
    #         headloss.loc[:,name] = abs(end_head - start_head)/link.length
    #     elif isinstance(link, Pump):
    #         # Negative of head gain for pumps 
    #         head_gain = -(end_head - start_head)
    #         head_gain[head_gain > 0] = 0
    #         headloss.loc[:,name] = head_gain
    #     else:
    #         # Total head loss for valves
    #         headloss.loc[:,name] = (end_head - start_head)
            
    #     # Headloss is 0 if the link is closed
    #     headloss.loc[results.link['status'].loc[:,name] == 0,name] = 0
        
    #results.link['headloss'] = headloss

def store_results_in_network(wn, m):
    """

    Parameters
    ----------
    wn: wntr.network.model.WaterNetworkModel
    m: wntr.sim.aml.aml.Model

    """
    mode = wn.options.hydraulic.demand_model
    for name, link in wn.links():
        if link._is_isolated:
            link._flow = 0
        else:
            link._flow = m.flow[name].value

    for name, link in wn.valves():
        link._setting = m.valve_setting[name].value

    for name, node in wn.junctions():
        if node._is_isolated:
            node._head = 0
            node._demand = 0
            node._pressure = 0
            node._leak_demand = 0
        else:
            node._head = m.head[name].value
            node._pressure = m.head[name].value - node.elevation
            if mode in ['PDD', 'PDA']:
                node._demand = m.demand[name].value
            else:
                node._demand = m.expected_demand[name].value
            if node.leak_status:
                node._leak_demand = m.leak_rate[name].value
            else:
                node._leak_demand = 0

    for name, node in wn.tanks():
        if node.leak_status:
            node._leak_demand = m.leak_rate[name].value
        else:
            node._leak_demand = 0
        node._demand = (sum(wn.get_link(link_name).flow for link_name in wn.get_links_for_node(name, 'INLET')) -
                       sum(wn.get_link(link_name).flow for link_name in wn.get_links_for_node(name, 'OUTLET')) -
                       node._leak_demand)

    for name, node in wn.reservoirs():
        node._head = node.head_timeseries.at(wn.sim_time)
        node._leak_demand = 0
        node._demand = (sum(wn.get_link(link_name).flow for link_name in wn.get_links_for_node(name, 'INLET')) -
                       sum(wn.get_link(link_name).flow for link_name in wn.get_links_for_node(name, 'OUTLET')))



import unittest
from os.path import abspath, dirname, join, exists

import wntr.epanet.exceptions

testdir = dirname(abspath(__file__))
datadir = join(testdir, "networks_for_testing")


class TestEpanetExceptions(unittest.TestCase):
    
    def test_epanet_exception(self):
        try:
            raise wntr.epanet.exceptions.EpanetException(213, '13:00:00 pm', 'Cannot specify am/pm for times greater than 12:00:00')
        except Exception as e:
            self.assertTupleEqual(e.args, ("(Error 213) invalid option value '13:00:00 pm' ['Cannot specify am/pm for times greater than 12:00:00']",))
        try:
            raise wntr.epanet.exceptions.EpanetException(999)
        except Exception as e:
            self.assertTupleEqual(e.args, ('(Error 999) unknown error',))
        try:
            raise wntr.epanet.exceptions.EpanetException(108)
        except Exception as e:
            self.assertTupleEqual(e.args, ('(Error 108) cannot use external file while hydraulics solver is active',))

    def test_epanet_syntax_error(self):
        try:
            raise wntr.epanet.exceptions.ENSyntaxError(223, line_num=38, line='I AM A SYNTAX ERROR')
        except SyntaxError as e:
            self.assertTupleEqual(e.args, ('(Error 223) not enough nodes in network, at line 38:\n   I AM A SYNTAX ERROR',))

    def test_epanet_key_error(self):
        try:
            raise wntr.epanet.exceptions.ENKeyError(206, 'NotACurve')
        except KeyError as e:
            self.assertTupleEqual(e.args, ("(Error 206) undefined curve, 'NotACurve'",))

    def test_epanet_value_error(self):
        try:
            raise wntr.epanet.exceptions.ENValueError(213, 423.0e28)
        except ValueError as e:
            self.assertTupleEqual(e.args, ('(Error 213) invalid option value 4.23e+30',))

    def test_broken_time_string(self):
        f = wntr.epanet.io.InpFile()
        inp_file = join(datadir, "bad_times.inp")
        try:
            f.read(inp_file)
        except Exception as e:
            self.assertIsInstance(e.__cause__, wntr.epanet.exceptions.ENValueError)
        else:
            self.fail('Failed to catch expected errors in bad_times.inp')

    def test_broken_syntax(self):
        f = wntr.epanet.io.InpFile()
        inp_file = join(datadir, "bad_syntax.inp")
        try:
            f.read(inp_file)
        except Exception as e:
            self.assertIsInstance(e, wntr.epanet.exceptions.ENSyntaxError)
        else:
            self.fail('Failed to catch expected errors in bad_syntax.inp')

    def test_bad_values(self):
        f = wntr.epanet.io.InpFile()
        inp_file = join(datadir, "bad_values.inp")
        try:
            f.read(inp_file)
        except Exception as e:
            self.assertIsInstance(e.__cause__, wntr.epanet.exceptions.ENKeyError)
        else:
            self.fail('Failed to catch expected errors in bad_values.inp')

if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
The wntr.msx.library module includes a library of multi-species water 
models

.. rubric:: Environment Variable

.. envvar:: WNTR_LIBRARY_PATH

    This environment variable, if set, will add additional folder(s) to the
    path to search for multi-species water quality model files, 
    (files with an ".msx", ".yaml", or ".json" file extension).
    Multiple folders should be separated using the "``;``" character.
    See :class:`~wntr.msx.library.ReactionLibrary` for more details.    
"""

from __future__ import annotations

import json
import logging
import os, sys
from typing import Any, ItemsView, Iterator, KeysView, List, Tuple, Union, ValuesView

if sys.version_info[0:2] <= (3, 11):
    from pkg_resources import resource_filename
else:
    from importlib.resources import files

from wntr.msx.base import ExpressionType, ReactionType, SpeciesType
from wntr.msx.model import MsxModel

try:
    import yaml

    yaml_err = None
except ImportError as e:
    yaml = None
    yaml_err = e

PIPE = ReactionType.PIPE
TANK = ReactionType.TANK
BULK = SpeciesType.BULK
WALL = SpeciesType.WALL
RATE = ExpressionType.RATE
EQUIL = ExpressionType.EQUIL
FORMULA = ExpressionType.FORMULA

logger = logging.getLogger(__name__)


class MsxLibrary:
    """Library of multi-species water quality models

    This object can be accessed and treated like a dictionary, where keys are
    the model names and the values are the model objects.

    Paths are added/processed in the following order:

    1. the builtin directory of reactions,
    2. any paths specified in the environment variable described below, with
       directories listed first having the highest priority,
    3. any extra paths specified in the constructor, searched in the order
    provided.

    Once created, the library paths cannot be modified. However, a model can
    be added to the library using :meth:`add_model_from_file` or
    :meth:`add_models_from_dir`.

    """

    def __init__(self, extra_paths: List[str] = None, include_builtins=True,
                 include_envvar_paths=True, load=True) -> None:
        """Library of multi-species water quality models

        Parameters
        ----------
        extra_paths : list of str, optional
            User-specified list of reaction library directories, by default
            None
        include_builtins : bool, optional
            Load files built-in with WNTR, by default True
        include_envvar_paths : bool, optional
            Load files from the paths specified in
            :envvar:`WNTR_LIBRARY_PATH`, by default True
        load : bool or str, optional
            Load the files immediately on creation, by default True.
            If a string, then it will be passed as the `duplicates` argument
            to the load function. See :meth:`reset_and_reload` for details.

        Raises
        ------
        TypeError
            If `extra_paths` is not a list
        """
        if extra_paths is None:
            extra_paths = list()
        elif not isinstance(extra_paths, (list, tuple)):
            raise TypeError("Expected a list or tuple, got {}".format(type(extra_paths)))

        self.__library_paths = list()

        self.__data = dict()

        if include_builtins:
            if sys.version_info[0:2] <= (3, 11):
                default_path = os.path.abspath(resource_filename(__name__, '.'))
            else:
                default_path = os.path.abspath(files('wntr.library.msx').joinpath('.'))
            if default_path not in self.__library_paths:
                self.__library_paths.append(default_path)

        if include_envvar_paths:
            environ_path = os.environ.get("WNTR_LIBRARY_PATH", None)
            if environ_path:
                lib_folders = environ_path.split(";")
                for folder in lib_folders:
                    if folder not in self.__library_paths:
                        self.__library_paths.append(os.path.abspath(folder))

        for folder in extra_paths:
            self.__library_paths.append(os.path.abspath(folder))
        if load:
            if isinstance(load, str):
                self.reset_and_reload(duplicates=load)
            else:
                self.reset_and_reload()

    def __repr__(self) -> str:
        if len(self.__library_paths) > 3:
            return "{}(initial_paths=[{}, ..., {}])".format(self.__class__.__name__, repr(self.__library_paths[0]), repr(self.__library_paths[-1]))
        return "{}({})".format(self.__class__.__name__, repr(self.__library_paths))

    def path_list(self) -> List[str]:
        """List of paths used to populate the library

        Returns
        -------
        list of str
            Copy of the paths used to **initially** populate the library
        """
        return self.__library_paths.copy()

    def reset_and_reload(self, duplicates: str = "error") -> List[Tuple[str, str, Any]]:
        """Load data from the configured directories into a library of models

        Note, this function is not recursive and does not 'walk' any of the
        library's directories to look for subfolders.

        The ``duplicates`` argument specifies how models that have the same name,
        or that have the same filename if a name isn't specified in the file,
        are handled. **Warning**, if two files in the same directory have models
        with the same name, there is no guarantee which will be read in first.

        The first directory processed is the builtin library data. Next, any
        paths specified in the environment variable are searched in the order listed
        in the variable. Finally, any directories specified by the user in the 
        constructor are processed in the order listed.

        Parameters
        ----------
        duplicates : {"error" | "skip" | "replace"}, optional
            by default ``"error"``

            - A value of of ``"error"`` raises an exception and stops execution. 
            - A value of ``"skip"`` will skip models with the same `name` as a model that already
              exists in the library.
            - A value of ``"replace"`` will replace any existing model with a model that is read 
              in that has the same `name`.

        Returns
        -------
        (filename, reason, obj) : (str, str, Any)
            the file not read in, the cause of the problem, and the object that was skipped/overwritten
            or the exception raised

        Raises
        ------
        TypeError
            If `duplicates` is not a string
        ValueError
            If `duplicates` is not a valid value
        IOError
            If `path_to_folder` is not a directory
        KeyError
            If `duplicates` is ``"error"`` and two models have the same name
        """
        if duplicates and not isinstance(duplicates, str):
            raise TypeError("The `duplicates` argument must be None or a string")
        elif duplicates.lower() not in ["error", "skip", "replace"]:
            raise ValueError('The `duplicates` argument must be None, "error", "skip", or "replace"')

        load_errors = list()
        for folder in self.__library_paths:
            errs = self.add_models_from_dir(folder, duplicates=duplicates)
            load_errors.extend(errs)
        return load_errors

    def add_models_from_dir(self, path_to_dir: str, duplicates: str = "error") -> List[Tuple[str, str, Union[MsxModel, Exception]]]:
        """Load all valid model files in a folder

        Note, this function is not recursive and does not 'walk' a directory
        tree.

        Parameters
        ----------
        path_to_dir : str
            Path to the folder to search
        duplicates : {"error", "skip", "replace"}, optional
            Specifies how models that have the same name, or that have the same
            filename if a name isn't specified in the file, are handled.
            **Warning**, if two files in the same directory have models with
            the same name, there is no guarantee which will be read in first,
            by default ``"error"``
            - A value of of ``"error"`` raises an exception and stops execution
            - A value of ``"skip"`` will skip models with the same `name` as a
              model that already exists in the library.
            - A value of ``"replace"`` will replace any existing model with a
              model that is read in that has the same `name`.

        Returns
        -------
        (filename, reason, obj) : tuple[str, str, Any]
            File not read in, the cause of the problem, and the object that was
            skipped/overwritten or the exception raised

        Raises
        ------
        TypeError
            If `duplicates` is not a string
        ValueError
            If `duplicates` is not a valid value
        IOError
            If `path_to_folder` is not a directory
        KeyError
            If `duplicates` is ``"error"`` and two models have the same name
        """
        if duplicates and not isinstance(duplicates, str):
            raise TypeError("The `duplicates` argument must be None or a string")
        elif duplicates.lower() not in ["error", "skip", "replace"]:
            raise ValueError('The `duplicates` argument must be None, "error", "skip", or "replace"')
        if not os.path.isdir(path_to_dir):
            raise IOError("The following path is not valid/not a folder, {}".format(path_to_dir))
        load_errors = list()
        folder = path_to_dir
        files = os.listdir(folder)
        for file in files:
            ext = os.path.splitext(file)[1]
            if ext is None or ext.lower() not in [".msx", ".json", ".yaml"]:
                continue
            if ext.lower() == ".msx":
                try:
                    new = MsxModel(file)
                except Exception as e:
                    logger.exception("Error reading file {}".format(os.path.join(folder, file)))
                    load_errors.append((os.path.join(folder, file), "load-failed", e))
                    continue
            elif ext.lower() == ".json":
                with open(os.path.join(folder, file), "r") as fin:
                    try:
                        new = MsxModel.from_dict(json.load(fin))
                    except Exception as e:
                        logger.exception("Error reading file {}".format(os.path.join(folder, file)))
                        load_errors.append((os.path.join(folder, file), "load-failed", e))
                        continue
            elif ext.lower() == ".yaml":
                if yaml is None:
                    logger.exception("Error reading file {}".format(os.path.join(folder, file)), exc_info=yaml_err)
                    load_errors.append((os.path.join(folder, file), "load-failed", yaml_err))
                    continue
                with open(os.path.join(folder, file), "r") as fin:
                    try:
                        new = MsxModel.from_dict(yaml.safe_load(fin))
                    except Exception as e:
                        logger.exception("Error reading file {}".format(os.path.join(folder, file)))
                        load_errors.append((os.path.join(folder, file), "load-failed", e))
                        continue
            else:  # pragma: no cover
                raise RuntimeError("This should be impossible to reach, since `ext` is checked above")
            new._orig_file = os.path.join(folder, file)
            if not new.name:
                new.name = os.path.splitext(os.path.split(file)[1])[0]
            if new.name not in self.__data:
                self.__data[new.name] = new
            else:  # this name exists in the library
                name = new.name
                if not duplicates or duplicates.lower() == "error":
                    raise KeyError('A model named "{}" already exists in the model; failed processing "{}"'.format(new.name, os.path.join(folder, file)))
                elif duplicates.lower() == "skip":
                    load_errors.append((new._orig_file, "skipped", new))
                    continue
                elif duplicates.lower() == "replace":
                    old = self.__data[name]
                    load_errors.append((old._orig_file, "replaced", old))
                    self.__data[name] = new
                else:  # pragma: no cover
                    raise RuntimeError("This should be impossible to get to, since `duplicates` is checked above")
        return load_errors

    def add_model_from_file(self, path_and_filename: str, name: str = None):
        """Load a model from a file and add it to the library

        Note, this **does not check** to see if a model exists with the same
        name, and it will automatically overwrite the existing model if one
        does exist.

        Parameters
        ----------
        path_to_file : str
            The full path and filename where the model is described.
        name : str, optional
            The name to use for the model instead of the name provided in the
            file or the filename, by default None
        """
        if not os.path.isfile(path_and_filename):
            raise IOError("The following path does not identify a file, {}".format(path_and_filename))

        ext = os.path.splitext(path_and_filename)[1]
        if ext is None or ext.lower() not in [".msx", ".json", ".yaml"]:
            raise IOError("The file is in an unknown format, {}".format(ext))
        if ext.lower() == ".msx":
            new = MsxModel(path_and_filename)
        elif ext.lower() == ".json":
            with open(path_and_filename, "r") as fin:
                new = MsxModel.from_dict(json.load(fin))
        elif ext.lower() == ".yaml":
            if yaml is None:
                raise RuntimeError("Unable to import yaml") from yaml_err
            with open(path_and_filename, "r") as fin:
                new = MsxModel.from_dict(yaml.safe_load(fin))
        else:  # pragma: no cover
            raise RuntimeError("This should be impossible to reach, since ext is checked above")
        new._orig_file = path_and_filename
        if not new.name:
            new.name = os.path.splitext(os.path.split(path_and_filename)[1])[0]
        if name is not None:
            new.name = name
        self.__data[new.name] = new

    def get_model(self, name: str) -> MsxModel:
        """Get a model from the library by model name

        Parameters
        ----------
        name : str
            Name of the model

        Returns
        -------
        MsxModel
            Model object
        """
        return self.__data[name]

    def model_name_list(self) -> List[str]:
        """Get the names of all models in the library
        
        Returns
        -------
        list of str
            list of model names
        """
        return list(self.keys())

    def __getitem__(self, __key: Any) -> Any:
        return self.__data.__getitem__(__key)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        return self.__data.__setitem__(__key, __value)

    def __delitem__(self, __key: Any) -> None:
        return self.__data.__delitem__(__key)

    def __contains__(self, __key: object) -> bool:
        return self.__data.__contains__(__key)

    def __iter__(self) -> Iterator:
        return self.__data.__iter__()

    def __len__(self) -> int:
        return self.__data.__len__()

    def keys(self) -> KeysView:
        return self.__data.keys()

    def items(self) -> ItemsView:
        return self.__data.items()

    def values(self) -> ValuesView:
        return self.__data.values()

    def clear(self) -> None:
        return self.__data.clear()



import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
from pandas.testing import assert_frame_equal, assert_series_equal

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "..", "..", "tests", "networks_for_testing")
netdir = join(testdir, "..", "..", "examples", "networks")


class TestScenario(unittest.TestCase):
    
    def test_distance_to_epicenter(self):
        inp_file = join(netdir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        earthquake = wntr.scenario.Earthquake((40, 55), 5, 10000.0)

        """
        Junction 12 is at (50,70)
        """
        R = earthquake.distance_to_epicenter(wn, wntr.network.Junction)
        expected = np.sqrt(np.power(50 - 40, 2) + np.power(70 - 55, 2))
        self.assertLess(np.abs(R["12"] - expected), 1e-5)

        wn = wntr.morph.scale_node_coordinates(wn, 100)
        R = earthquake.distance_to_epicenter(wn, wntr.network.Junction)
        expected = np.sqrt(np.power(5000 - 40, 2) + np.power(7000 - 55, 2))
        self.assertLess(np.abs(R["12"] - expected), 1e-5)

    def test_correction_factor(self):
        earthquake = wntr.scenario.Earthquake((40, 55), 5, 10000.0)

        data = pd.DataFrame(
            columns=["Diameter", "Material", "Topography", "Liquifaction"],
            index=["P1", "P2", "P3"],
        )
        data.loc["P1", :] = ["Small", "ACP", "Narrow valley", "Total"]
        data.loc["P2", :] = ["Medium", "PV", "Terrace", "Partial"]
        data.loc["P3", :] = ["Large", "SP", "Stiff alluvial", "None"]
        C = earthquake.correction_factor(data)
        expected = pd.Series(
            [1 * 1.2 * 3.2 * 2.4, 0.8 * 1 * 1.5 * 2, 0.5 * 0.3 * 0.4 * 1],
            index=["P1", "P2", "P3"],
        )
        assert_series_equal(C, expected)

    def test_pga_attenuation(self):
        R = pd.Series({"1": 1000.0})

        earthquake = wntr.scenario.Earthquake((0, 0), 5, 10000.0)

        pga = earthquake.pga_attenuation_model(R, method=1)
        # print(pga['1'])
        self.assertLess(np.abs(pga["1"] - 1.3275e-1), 1e-5)

        pga = earthquake.pga_attenuation_model(R, method=2)
        # print(pga['1'])
        self.assertLess(np.abs(pga["1"] - 9.6638e-2), 1e-6)

        pga = earthquake.pga_attenuation_model(R, method=3)
        # print(pga['1'])
        self.assertLess(np.abs(pga["1"] - 1.2789e-3), 1e-7)

        pga = earthquake.pga_attenuation_model(R)
        # print(pga['1'])
        self.assertLess(np.abs(pga["1"] - 7.6887e-2), 1e-6)

    def test_pgv_attenuation(self):
        R = pd.Series({"1": 1000})

        # Yu and Jin, 2008
        earthquake = wntr.scenario.Earthquake((0, 0), 5, 0)
        pgv = earthquake.pgv_attenuation_model(R, method=1)
        self.assertLess(np.abs(pgv["1"] - 0.0531), 0.0001)

        earthquake = wntr.scenario.Earthquake((0, 0), 7, 0)
        pgv = earthquake.pgv_attenuation_model(R, method=1)
        self.assertLess(np.abs(pgv["1"] - 1.8829), 0.0001)

        earthquake = wntr.scenario.Earthquake((0, 0), 5, 0)
        pgv = earthquake.pgv_attenuation_model(R, method=2)
        self.assertLess(np.abs(pgv["1"] - 0.0884), 0.0001)

        earthquake = wntr.scenario.Earthquake((0, 0), 7, 0)
        pgv = earthquake.pgv_attenuation_model(R, method=2)
        self.assertLess(np.abs(pgv["1"] - 2.3361), 0.0001)

        earthquake = wntr.scenario.Earthquake((0, 0), 5, 0)
        pgv = earthquake.pgv_attenuation_model(R)
        self.assertLess(np.abs(pgv["1"] - 0.0707), 0.0001)

        earthquake = wntr.scenario.Earthquake((0, 0), 7, 0)
        pgv = earthquake.pgv_attenuation_model(R)
        self.assertLess(np.abs(pgv["1"] - 2.1095), 0.0001)

    def test_repair_rate(self):

        PGV = pd.Series({"1": 0.0531})
        C = pd.Series({"1": 0.5})
        earthquake = wntr.scenario.Earthquake((0, 0), 0, 0)

        RR = earthquake.repair_rate_model(PGV, C=1, method=1)
        self.assertLess(np.abs(RR["1"] - 1.2823e-5), 1e-9)

        RR = earthquake.repair_rate_model(PGV, C, method=1)
        self.assertLess(np.abs(RR["1"] - 6.4113e-6), 1e-10)

        RR = earthquake.repair_rate_model(PGV, C=1, method=2)
        self.assertLess(np.abs(RR["1"] - 8.4132e-6), 1e-10)

        RR = earthquake.repair_rate_model(PGV, C, method=2)
        self.assertLess(np.abs(RR["1"] - 4.2066e-6), 1e-10)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
The wntr.msx.io module includes functions that convert the MSX reaction
model to other data formats, create an MSX model from a file, and write 
the MSX model to a file.
"""
import logging
import json


logger = logging.getLogger(__name__)

def to_dict(msx) -> dict:
    """
    Convert a MsxModel into a dictionary
    
    Parameters
    ----------
    msx : MsxModel
        The MSX reaction model.

    Returns
    -------
    dict
        Dictionary representation of the MsxModel
    """
    return msx.to_dict()

def from_dict(d: dict):
    """
    Create or append a MsxModel from a dictionary

    Parameters
    ----------
    d : dict
        Dictionary representation of the water network model.

    Returns
    -------
    MsxModel
    
    """
    from wntr.msx.model import MsxModel
    return MsxModel.from_dict(d)

def write_json(msx, path_or_buf, as_library=False, indent=4, **kw_json):
    """
    Write the MSX model to a JSON file

    Parameters
    ----------
    msx : MsxModel
        The model to output.
    path_or_buf : str or IO stream
        Name of the file or file pointer.
    as_library : bool, optional
        Strip out network-specific elements if True, by default False.
    kw_json : keyword arguments
        Arguments to pass directly to :meth:`json.dump`.
    """
    d = to_dict(msx)
    if as_library:
        d.get('network_data', {}).get('initial_quality',{}).clear()
        d.get('network_data', {}).get('parameter_values',{}).clear()
        d.get('network_data', {}).get('sources',{}).clear()
        d.get('network_data', {}).get('patterns',{}).clear()
        d.get('options', {}).get('report',{}).setdefault('nodes', None)
        d.get('options', {}).get('report',{}).setdefault('links', None)
    if isinstance(path_or_buf, str):
        with open(path_or_buf, "w") as fout:
            json.dump(d, fout, indent=indent, **kw_json)
    else:
        json.dump(d, path_or_buf, indent=indent, **kw_json)

def read_json(path_or_buf, **kw_json):
    """
    Create or append a WaterNetworkModel from a JSON file

    Parameters
    ----------
    f : str
        Name of the file or file pointer.
    kw_json : keyword arguments
        Keyword arguments to pass to `json.load`.

    Returns
    -------
    MsxModel
    
    """
    if isinstance(path_or_buf, str):
        with open(path_or_buf, "r") as fin:
            d = json.load(fin, **kw_json)
    else:
        d = json.load(path_or_buf, **kw_json)
    return from_dict(d)

def write_msxfile(msx, filename):
    """
    Write an EPANET-MSX input file (.msx)

    Parameters
    ----------
    msx : MsxModel
        The model to write
    filename : str
        The filename to use for output
    """
    from wntr.epanet.msx.io import MsxFile
    MsxFile.write(filename, msx)

def read_msxfile(filename, append=None):
    """
    Read in an EPANET-MSX input file (.msx)
    
    Parameters
    ----------
    filename : str
        The filename to read in.
    append : MsxModel
        An existing model to add data into, by default None.
    """
    from wntr.epanet.msx.io import MsxFile
    return MsxFile.read(filename, append)



# coding: utf-8
"""Exceptions for EPANET toolkit and IO operations."""

from typing import List

EN_ERROR_CODES = {
    # Runtime errors
    1: "At %s, system hydraulically unbalanced - convergence to a hydraulic solution was not achieved in the allowed number of trials",
    2: "At %s, system may be hydraulically unstable - hydraulic convergence was only achieved after the status of all links was held fixed",
    3: "At %s, system disconnected - one or more nodes with positive demands were disconnected for all supply sources",
    4: "At %s, pumps cannot deliver enough flow or head - one or more pumps were forced to either shut down (due to insufficient head) or operate beyond the maximum rated flow",
    5: "At %s, valves cannot deliver enough flow - one or more flow control valves could not deliver the required flow even when fully open",
    6: "At %s, system has negative pressures - negative pressures occurred at one or more junctions with positive demand",
    101: "insufficient memory available",
    102: "no network data available",
    103: "hydraulics not initialized",
    104: "no hydraulics for water quality analysis",
    105: "water quality not initialized",
    106: "no results saved to report on",
    107: "hydraulics supplied from external file",
    108: "cannot use external file while hydraulics solver is active",
    109: "cannot change time parameter when solver is active",
    110: "cannot solve network hydraulic equations",
    120: "cannot solve water quality transport equations",
    # Apply only to an input file
    200: "one or more errors in input file %s",
    201: "syntax error (%s)",
    # Apply to both IO file and API functions
    202: "illegal numeric value, %s",
    203: "undefined node, %s",
    204: "undefined link, %s",
    205: "undefined time pattern, %s",
    206: "undefined curve, %s",
    207: "attempt to control a CV/GPV link",
    208: "illegal PDA pressure limits",
    209: "illegal node property value",
    211: "illegal link property value",
    212: "undefined trace node",
    213: "invalid option value %s",
    214: "too many characters in input line",
    215: "duplicate ID label",
    216: "reference to undefined pump",
    217: "pump has no head curve or power defined",
    219: "illegal valve connection to tank node",
    220: "illegal valve connection to another valve",
    221: "misplaced rule clause in rule-based control",
    222: "link assigned same start and end nodes",
    # Network consistency
    223: "not enough nodes in network",
    224: "no tanks or reservoirs in network",
    225: "invalid lower/upper levels for tank",
    226: "no head curve or power rating for pump",
    227: "invalid head curve for pump",
    230: "nonincreasing x-values for curve",
    233: "network has unconnected node",
    234: "network has an unconnected node with ID %s",
    # API functions only
    240: "nonexistent water quality source",
    241: "nonexistent control",
    250: "invalid format (e.g. too long an ID name)",
    251: "invalid parameter code",
    252: "invalid ID name",
    253: "nonexistent demand category",
    254: "node with no coordinates",
    255: "invalid link vertex",
    257: "nonexistent rule",
    258: "nonexistent rule clause",
    259: "attempt to delete a node that still has links connected to it",
    260: "attempt to delete node assigned as a Trace Node",
    261: "attempt to delete a node or link contained in a control",
    262: "attempt to modify network structure while a solver is open",
    263: "node is not a tank",
    # File errors
    301: "identical file names used for different types of files",
    302: "cannot open input file %s",
    303: "cannot open report file %s",
    304: "cannot open binary output file %s",
    305: "cannot open hydraulics file %s",
    306: "hydraulics file does not match network data",
    307: "cannot read hydraulics file %s",
    308: "cannot save results to binary file %s",
    309: "cannot save results to report file %s",
}
"""A dictionary of the error codes and their meanings from the EPANET toolkit.

:meta hide-value:
"""


class EpanetException(Exception):

    def __init__(self, code: int, *args: List[object], line_num=None, line=None) -> None:
        """An Exception class for EPANET Toolkit and IO exceptions.

        Parameters
        ----------
        code : int
            The EPANET error code
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these will be used to
            replace the format, otherwise they will be output at the end of the Exception message.
        line_num : int, optional
            The line number, if reading an INP file, by default None
        line : str, optional
            The contents of the line, by default None
        """
        msg = EN_ERROR_CODES.get(code, "unknown error")
        if args is not None:
            args = [*args]
        if r"%" in msg and len(args) > 0:
            msg = msg % repr(args.pop(0))
        if len(args) > 0:
            msg = msg + " " + repr(args)
        if line_num:
            msg = msg + ", at line {}".format(line_num)
        if line:
            msg = msg + ":\n   " + str(line)
        msg = "(Error {}) ".format(code) + msg
        super().__init__(msg)


class ENSyntaxError(EpanetException, SyntaxError):
    def __init__(self, code, *args, line_num=None, line=None) -> None:
        """An EPANET exception class that also subclasses SyntaxError

        Parameters
        ----------
        code : int
            The EPANET error code
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these will be used to
            replace the format, otherwise they will be output at the end of the Exception message.
        line_num : int, optional
            The line number, if reading an INP file, by default None
        line : str, optional
            The contents of the line, by default None
        """
        super().__init__(code, *args, line_num=line_num, line=line)


class ENKeyError(EpanetException, KeyError):
    def __init__(self, code, name, *args, line_num=None, line=None) -> None:
        """An EPANET exception class that also subclasses KeyError.

        Parameters
        ----------
        code : int
            The EPANET error code
        name : str
            The key/name/id that is missing
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these will be used to
            replace the format, otherwise they will be output at the end of the Exception message.
        line_num : int, optional
            The line number, if reading an INP file, by default None
        line : str, optional
            The contents of the line, by default None
        """

        super().__init__(code, name, *args, line_num=line_num, line=line)


class ENValueError(EpanetException, ValueError):
    def __init__(self, code, value, *args, line_num=None, line=None) -> None:
        """An EPANET exception class that also subclasses ValueError

        Parameters
        ----------
        code : int
            The EPANET error code
        value : Any
            The value that is invalid
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these will be used to
            replace the format, otherwise they will be output at the end of the Exception message.
        line_num : int, optional
            The line number, if reading an INP file, by default None
        line : str, optional
            The contents of the line, by default None
        """
        super().__init__(code, value, *args, line_num=line_num, line=line)
